#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""主体、斑块之间可以产生连接。

Actor, PatchCell can be used to create links.
"""


from __future__ import annotations

import contextlib
from abc import abstractmethod
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    overload,
)

import numpy as np
import pandas as pd
from loguru import logger

with contextlib.suppress(ImportError):
    import networkx as nx
try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from abses._bases.errors import ABSESpyError
from abses.sequences import ActorsList
from abses.tools.func import make_list
from abses.tools.viz import get_marker

if TYPE_CHECKING:
    from abses import MainModel
    from abses.actor import Actor
    from abses.cells import PatchCell
    from abses.container import UniqueID, _CellAgentsContainer
    from abses.sequences import Link

LinkingNode: TypeAlias = "Actor | PatchCell"
Direction: TypeAlias = Optional[Literal["in", "out"]]
DEFAULT_TARGETS: Tuple[str, str] = ("cell", "actor")
TargetName: TypeAlias = Union[Literal["cell", "actor", "self"], str]
AttrGetter: TypeAlias = Union["Link", ActorsList["Link"]]


def get_node_unique_id(node: Any) -> UniqueID:
    """When import actors from graph, decide the unique ID."""
    if not isinstance(node, (str, int)):
        logger.warning(
            f"Using repr for '{type(node)}' unique ID to create actor."
        )
        return repr(node)
    return node


class _LinkContainer:
    """Container for links."""

    def __init__(self) -> None:
        self._back_links: Dict[str, Dict[LinkingNode, Set]] = {}
        self._links: Dict[str, Dict[LinkingNode, Set]] = {}
        self._cached_networks: Dict[str, object] = {}

    @property
    def links(self) -> Tuple[str, ...]:
        """Get the links of a certain type."""
        return tuple(self._links.keys())

    def _add_a_link_name(self, link_name: str) -> None:
        """Add a link."""
        self._links[link_name] = {}
        self._back_links[link_name] = {}

    def owns_links(
        self, node: LinkingNode, direction: Direction = "out"
    ) -> Tuple[str, ...]:
        """The links a specific node owns."""
        if direction == "out":
            data = self._links
        elif direction == "in":
            data = self._back_links
        elif direction is None:
            links_in = self.owns_links(node, direction="in")
            links_out = self.owns_links(node, direction="out")
            return tuple(set(links_in) | set(links_out))
        else:
            raise ValueError(f"Invalid direction '{direction}'.")
        links = {link for link, agents in data.items() if node in agents}
        return tuple(links)

    @overload
    def get_graph(
        self, link_name: str, directions: bool = False
    ) -> "nx.Graph":
        ...

    @overload
    def get_graph(
        self, link_name: str, directions: bool = True
    ) -> "nx.DiGraph":
        ...

    def get_graph(
        self, link_name: str, directions: bool = False
    ) -> "nx.Graph | nx.DiGraph":
        """Get the networkx graph.

        Parameters:
            link_name:
                The link name for converting into a graph.

        Raises:
            ImportError:
                If networkx is not installed.
        """
        if "nx" not in globals():
            raise ImportError(
                "You need to install networkx to use this function."
            )
        creating_using = nx.DiGraph if directions else nx.Graph
        graph = nx.from_dict_of_lists(self._links[link_name], creating_using)
        self._cached_networks[link_name] = graph
        return graph

    def _register_link(
        self, link_name: str, source: LinkingNode, target: LinkingNode
    ) -> None:
        """Register a link."""
        if link_name not in self._links:
            self._add_a_link_name(link_name)
        if source not in self._links[link_name]:
            self._links[link_name][source] = set()
        if target not in self._back_links[link_name]:
            self._back_links[link_name][target] = set()

    def has_link(
        self, link_name: str, source: LinkingNode, target: LinkingNode
    ) -> Tuple[bool, bool]:
        """If any link exists between source and target.

        Parameters:
            link_name:
                The name of the link.
            source:
                The source node.
            target:
                The target node.

        Raises:
            KeyError:
                If the link name does not exist.

        Returns:
            tuple:
                A tuple of two booleans.
                The first element is True if the link exists from source to target.
                The second element is True if the link exists from target to source.
        """
        if link_name not in self.links:
            raise KeyError("No link named {link_name}.")
        data = self._links[link_name]
        to = False if source not in data else target in data.get(source, [])
        by = False if target not in data else source in data.get(target, [])
        return to, by

    def add_a_link(
        self,
        link_name: str,
        source: LinkingNode,
        target: LinkingNode,
        mutual: bool = False,
    ) -> None:
        """Add a link from source to target.

        Parameters:
            link_name:
                The name of the link.
            source:
                The source node.
            target:
                The target node.
            mutual:
                If the link is mutual.
        """
        self._register_link(link_name, source, target)
        self._links[link_name][source].add(target)
        self._back_links[link_name][target].add(source)
        if mutual:
            self.add_a_link(
                link_name, target=source, source=target, mutual=False
            )

    def remove_a_link(
        self,
        link_name: str,
        source: LinkingNode,
        target: LinkingNode,
        mutual: bool = False,
    ) -> None:
        """Remove a specific link.

        Parameters:
            link_name:
                The name of the link.
            source:
                The source node.
            target:
                The target node.
            mutual:
                If delete the link mutually.

        Raises:
            ABSESpyError:
                If the link from source to target does not exist.
        """
        if not self.has_link(link_name, source, target)[0]:
            raise ABSESpyError(f"Link from {source} to {target} not found.")
        self._links[link_name].get(source, set()).remove(target)
        self._back_links[link_name].get(target, set()).remove(source)
        if mutual:
            self.remove_a_link(
                link_name, target=source, source=target, mutual=False
            )

    def _clean_link_name(
        self, link_name: Optional[str | Iterable[str]]
    ) -> List[str]:
        """Clean the link name."""
        if link_name is None:
            link_name = self.links
        if isinstance(link_name, str):
            link_name = [link_name]
        if not isinstance(link_name, Iterable):
            raise TypeError(f"{link_name} is not an iterable.")
        return list(link_name)

    def clean_links_of(
        self,
        node: LinkingNode,
        link_name: Optional[str] = None,
        direction: Direction = None,
    ) -> None:
        """Clean the links of a node.

        Parameters:
            node:
                The node to clean the links.
            link_name:
                The name of the link to clean.
                If None, clean all related links for the node.
            direction:
                The direction of the link ('in' or 'out').
                If None, clean both directions (both out links and in links).

        Raises:
            ValueError:
                If the direction is not 'in' or 'out'.
        """
        if direction == "in":
            data = self._back_links
            another_data = self._links
        elif direction == "out":
            data = self._links
            another_data = self._back_links
        elif direction is None:
            self.clean_links_of(node, link_name, direction="in")
            self.clean_links_of(node, link_name, direction="out")
            return
        else:
            raise ValueError(
                f"Invalid direction {direction}, please choose from 'in' or 'out'."
            )
        for name in self._clean_link_name(link_name):
            to_clean = data[name].pop(node, set())
            for another_node in to_clean:
                another_data[name][another_node].remove(node)

    def linked(
        self,
        node: LinkingNode,
        link_name: Optional[str] = None,
        direction: Direction = None,
        default: bool = False,
    ) -> Set[LinkingNode]:
        """Get the linked nodes.

        Parameters:
            node:
                The node to get the linked nodes.
            link_name:
                The name of the link.
                If None, get all type of links.
            direction:
                The direction of the link ('in' or 'out').

        Raises:
            ValueError:
                If the direction is not 'in' or 'out'.

        Returns:
            The linked Actors or PatchCells with the input node.
        """
        link_names = self._clean_link_name(link_name=link_name)
        if direction == "in":
            data = self._back_links
        elif direction == "out":
            data = self._links
        elif direction is None:
            in_links = self.linked(node, link_name, direction="in")
            out_links = self.linked(node, link_name, direction="out")
            return in_links | out_links
        else:
            raise ValueError(f"Invalid direction {direction}")
        agents: Set[LinkingNode] = set()
        for name in link_names:
            if name not in data and default:
                continue
            agents = agents.union(data[name].get(node, set()))
        return agents

    def _check_is_node(
        self,
        node: UniqueID | LinkingNode,
        mapping_dict: Optional[Dict[UniqueID, Actor]] = None,
    ) -> LinkingNode:
        if mapping_dict:
            unique_id = get_node_unique_id(node)
            node = mapping_dict[unique_id]
        if not isinstance(node, _LinkNode):
            raise TypeError(
                f"Invalid node type {type(node)}, mapping: {mapping_dict}."
            )
        return node

    def add_links_from_graph(
        self,
        graph: "nx.Graph",
        link_name: str,
        mapping_dict: Optional[Dict[UniqueID, Actor]] = None,
        mutual: Optional[bool] = None,
    ) -> None:
        """Add links from graph."""
        if mutual is None:
            mutual = False if isinstance(graph, nx.DiGraph) else True
        if mapping_dict is None:
            mapping_dict = {}
        edges = 0
        for source, targets in nx.to_dict_of_lists(graph).items():
            source = self._check_is_node(source, mapping_dict)
            for target in targets:
                target = self._check_is_node(target, mapping_dict)
                self.add_a_link(link_name, source, target, mutual=mutual)
                edges += 1
        logger.info(f"Imported links {edges} links from graph {graph}.")


class _LinkProxy:
    """Proxy for linking."""

    def __init__(self, node: LinkingNode, model: MainModel) -> None:
        self.node: LinkingNode = node
        self.model: MainModel = model
        self.human: _LinkContainer = model.human

    def __contains__(self, link_name: str) -> bool:
        """Check if the link exists."""
        return link_name in self.human.links

    def __eq__(self, __value: object) -> bool:
        """Check if the links are equal to a set of strings."""
        if not isinstance(__value, Iterable):
            return NotImplemented
        return set(__value) == set(self.owning())

    def __repr__(self) -> str:
        return str(self.owning())

    def owning(self, direction: Direction = None) -> Tuple[str, ...]:
        """Links that this object has.

        Parameters:
            direction:
                The direction of the link ('in' or 'out').
                If None, return both out links and in links.

        Returns:
            The links that this object has.
        """
        return self.human.owns_links(self.node, direction=direction)

    def get(
        self,
        link_name: Optional[str] = None,
        direction: Direction = "out",
        default: bool = False,
    ) -> ActorsList[LinkingNode]:
        """Get the linked nodes."""
        agents = self.human.linked(
            self.node, link_name, direction=direction, default=default
        )
        return ActorsList(self.model, agents)

    def has(
        self, link_name: str, node: Optional[LinkingNode] = None
    ) -> Tuple[bool, bool]:
        """Check if the node has the link.

        Parameters:
            link_name:
                The name of the link.
            node:
                The node to check if it has the link with the current node.
                If None, check if the current node has any link.

        Returns:
            tuple:
                A tuple of two booleans.
                The first element is True if the link exists from me to other.
                The second element is True if the link exists from other to me.
        """
        if node is None:
            has_in = link_name in self.owning("in")
            has_out = link_name in self.owning("out")
            return has_out, has_in
        return self.human.has_link(link_name, self.node, node)

    def to(
        self, node: LinkingNode, link_name: str, mutual: bool = False
    ) -> None:
        """Link to another node.

        Parameters:
            node:
                The node to link to.
            link_name:
                The name of the link.
            mutual:
                If the link is mutual. Defaults to False.
        """
        self.human.add_a_link(
            link_name=link_name, source=self.node, target=node, mutual=mutual
        )

    def by(
        self, node: LinkingNode, link_name: str, mutual: bool = False
    ) -> None:
        """Make this node linked by another node.

        Parameters:
            node:
                The node to link by.
            link_name:
                The name of the link.
            mutual:
                If the link is mutual. Defaults to False.
        """
        self.human.add_a_link(
            link_name=link_name, source=node, target=self.node, mutual=mutual
        )

    def unlink(self, node: LinkingNode, link_name: str, mutual: bool = False):
        """Remove the link between me and another node.

        Parameters:
            node:
                The node to unlink with.
            link_name:
                The name of the link.
            mutual:
                If delete link mutually. Defaults to False.

        Raises:
            ABSESpyError:
                If the link from source to target does not exist.
        """
        self.human.remove_a_link(
            link_name=link_name, source=self.node, target=node, mutual=mutual
        )

    def clean(
        self, link_name: Optional[str] = None, direction: Direction = None
    ):
        """Clean all the related links from this node.

        Parameters:
            link_name:
                The name of the link.
                If None, clean all related links for the node.
            direction:
                The direction of the link ('in' or 'out').
                If None, clean both directions (both out links and in links).

        Raises:
            ValueError:
                If the direction is not 'in' or 'out'.
        """
        self.human.clean_links_of(
            self.node, link_name=link_name, direction=direction
        )


class _BreedDescriptor:
    """A descriptor to get the breed of a node."""

    def __get__(self, _: Any, owner: Any) -> str:
        return owner.__name__ if owner else self.__class__.__name__


class _LinkNode:
    """节点类"""

    unique_id: int = -1
    breed = _BreedDescriptor()

    @abstractmethod
    def _default_redirection(self, target: Optional[TargetName]) -> AttrGetter:
        """默认重定向"""

    @classmethod
    def viz_attrs(
        cls, render_marker: bool = False, **kwargs
    ) -> Dict[str, Any]:
        """Return the attributes for viz."""
        maker = getattr(cls, "marker", "o")
        return {
            "marker": get_marker(maker) if render_marker else maker,
            "color": getattr(cls, "color", "black"),
            "alpha": getattr(cls, "alpha", 1.0),
        } | kwargs

    @cached_property
    def link(self) -> _LinkProxy:
        """A proxy which can be used to manipulate the links:

        1. `link.to()`: creates a new link from this actor to another.
        2. `link.by()`: creates a new link from another to this actor.
        3. `link.get()`: gets the links of this actor.
        4. `link.has()`: checks if there is a link between this actor and another.
        5. `link.unlink()`: removes a link between this actor and another.
        6. `link.clean()`: removes all links of this actor.
        """
        return _LinkProxy(cast(LinkingNode, self), getattr(self, "model"))

    def _redirect_getting(self, target: Optional[TargetName]) -> AttrGetter:
        """Which targets should be used when getting or setting."""
        # If the target is not None, get the default target.
        if target is None or target in DEFAULT_TARGETS:
            return self._default_redirection(target)
        if target in self.link:
            return self.link.get(link_name=target)
        raise ABSESpyError(f"Unknown target {target}.")

    def get(
        self,
        attr: str,
        target: Optional[TargetName] = None,
    ) -> Any:
        """Gets attribute value from target.

        Parameters:
            attr:
                The name of the attribute to get.
            target:
                The target to get the attribute from.
                If None, the agent itself is the target.
                If the target is an agent, get the attribute from the agent.
                If the target is a cell, get the attribute from the cell.

        Returns:
            The value of the attribute.
        """
        if target in DEFAULT_TARGETS:
            return self._default_redirection(target).get(attr, "self")
        if target == "self":
            return getattr(self, attr)
        if hasattr(self, attr):
            assert (
                target is None
            ), f"The target '{target}' is set when '{self}' already has attr '{attr}'."
            return getattr(self, attr)
        return self._redirect_getting(target=target).get(attr, "self")

    def set(
        self, attr: str, value: Any, target: Optional[TargetName] = None
    ) -> None:
        """Sets the value of an attribute.

        Parameters:
            attr:
                The name of the attribute to set.
            value:
                The value to set the attribute to.
            target:
                The target to set the attribute on. If None, the agent itself is the target.
                1. If the target is an agent, set the attribute on the agent.
                2. If the target is a cell, set the attribute on the cell.

        Raises:
            TypeError:
                If the attribute is not a string.
            ABSESpyError:
                If the attribute is protected.
        """
        # If the attribute is not a string, raise an error.
        if not isinstance(attr, str):
            raise TypeError("The attribute must be a string.")
        # If the attribute is protected, raise an error
        if attr.startswith("_"):
            raise ABSESpyError(f"Attribute '{attr}' is protected.")
        if target in DEFAULT_TARGETS:
            self._default_redirection(target).set(attr, value, "self")
            return
        if target == "self":
            if not hasattr(self, attr):
                raise AttributeError(f"{self} has no attribute '{attr}'.")
            setattr(self, attr, value)
            return
        # Set the attribute on the target.
        if hasattr(self, attr):
            assert (
                target is None
            ), f"The target '{target}' is set when '{self}' already has attr '{attr}'."
            setattr(self, attr, value)
        else:
            new_target = self._redirect_getting(target=target)
            new_target.set(attr, value, "self")

    def summary(
        self,
        coords: bool = False,
        attrs: Optional[Iterable[str] | str] = None,
    ) -> pd.Series:
        """Returns a summary of the object."""
        geo_type = self.get("geo_type")
        if geo_type in ("Point", "Cell"):
            a, b = self.get("coordinate" if coords else "pos")
        else:
            a, b = np.nan, np.nan
        result = {
            "breed": self.breed,
            "geo_type": geo_type,
            "x" if coords else "row": a,
            "y" if coords else "col": b,
        }
        result.update({attr: self.get(attr) for attr in make_list(attrs)})
        return pd.Series(result, name=self.unique_id)


class _LinkNodeCell(_LinkNode):
    def _default_redirection(
        self, target: Optional[TargetName]
    ) -> _CellAgentsContainer | _LinkNodeCell:
        if target == "cell":
            return self
        return cast(_CellAgentsContainer, getattr(self, "agents"))


class _LinkNodeActor(_LinkNode):
    def _default_redirection(
        self, target: Optional[TargetName]
    ) -> _LinkNodeActor | Optional[_LinkNodeCell]:
        if target == "actor":
            return self
        return cast(Optional[_LinkNodeCell], getattr(self, "at"))



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
The spatial module.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    Optional,
    Protocol,
    Type,
    overload,
)

import numpy as np
import pyproj
import rasterio
import xarray as xr
from mesa.space import Coordinate

from abses._bases.modules import CompositeModule, HowCreation
from abses.cells import PatchCell
from abses.patch import (
    CRS,
    CellFilter,
    PatchModule,
    Raster,
    _PatchModuleFactory,
)
from abses.random import ListRandom
from abses.sequences import ActorsList
from abses.viz.viz_nature import _VizNature

if TYPE_CHECKING:
    from abses.main import MainModel


class _PatchModuleProtocol(Protocol):
    @property
    def cells(self) -> ActorsList[PatchCell]:
        """The cells stored in this layer."""

    @property
    def mask(self) -> np.ndarray:
        """Where is not accessible."""

    @property
    def cell_properties(self) -> set[str]:
        """The accessible attributes of cells stored in this layer.
        All `PatchCell` methods decorated by `raster_attribute` should be appeared here.
        """

    @property
    def xda(self) -> xr.DataArray:
        """Get the xarray raster layer with spatial coordinates."""

    @property
    def plot(self) -> _VizNature:
        """Plotting"""

    @property
    def attributes(self) -> set[str]:
        """All accessible attributes from this layer."""

    @property
    def shape2d(self) -> Coordinate:
        """Raster shape in 2D (height, width).
        This is useful when working with 2d `numpy.array`.
        """

    @property
    def shape3d(self) -> Coordinate:
        """Raster shape in 3D (1, heigh, width).
        This is useful when working with `rasterio` band.
        """

    @property
    def array_cells(self) -> np.ndarray:
        """Array type of the `PatchCell` stored in this module."""

    @property
    def coords(self) -> Coordinate:
        """Coordinate system of the raster data.
        This is useful when working with `xarray.DataArray`.
        """

    def to_crs(self, crs, inplace=False) -> Optional[PatchModule]:
        """Converting the raster data to a another CRS."""

    def dynamic_var(self, attr_name: str) -> np.ndarray:
        """Update and get dynamic variable.

        Parameters:
            attr_name:
                The dynamic variable to retrieve.

        Returns:
            2D numpy.ndarray data of the variable.
        """

    def get_rasterio(
        self, attr_name: str | None = None
    ) -> rasterio.MemoryFile:
        """Gets the Rasterio raster layer corresponding to the attribute. Save to a temporary rasterio memory file.

        Parameters:
            attr_name:
                The attribute name for creating the rasterio file.

        Returns:
            The rasterio tmp memory file of raster.
        """

    def get_xarray(self, attr_name: Optional[str] = None) -> xr.DataArray:
        """Get the xarray raster layer with spatial coordinates.

        Parameters:
            attr_name:
                The attribute to retrieve. If None (by default),
                return all available attributes (3D DataArray).
                Otherwise, 2D DataArray of the chosen attribute.

        Returns:
            Xarray.DataArray data with spatial coordinates of the chosen attribute.
        """

    def out_of_bounds(self, pos: Coordinate) -> bool:
        """Check if a pos is out of the bounds.
        This is a protocol method.
        """

    @property
    def random(self) -> ListRandom:
        """Randomly"""

    def select(
        self,
        where: Optional[CellFilter] = None,
    ) -> ActorsList[PatchCell]:
        """Select cells from this layer.
        Also has a shortcut alias for this method: `.sel`.

        Parameters:
            where:
                The condition to select cells.
                If None (by default), select all cells.
                If a string, select cells by the attribute name.
                If a numpy.ndarray, select cells by the mask array.
                If a Shapely Geometry, select cells by the intersection with the geometry.

        Raises:
            TypeError:
                If the input type is not supported.

        Returns:
            An `ActorsList` with all selected cells stored.
        """

    def apply(
        self, ufunc: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> np.ndarray:
        """Apply a function to array cells.

        Parameters:
            ufunc:
                A function to apply.
            *args:
                Positional arguments to pass to the function.
            **kwargs:
                Keyword arguments to pass to the function.

        Returns:
            The result of the function applied to the array cells.
        """

    def coord_iter(self) -> Iterator[tuple[Coordinate, PatchCell]]:
        """
        An iterator that returns coordinates as well as cell contents.
        """

    def apply_raster(
        self, data: Raster, attr_name: Optional[str], **kwargs: Any
    ) -> None:
        """Apply raster data to the cells.

        Parameters:
            data:
                np.ndarray data: 2D numpy array with shape (1, height, width).
                xr.DataArray data: xarray DataArray with spatial coordinates.
                xr.Dataset data: xarray Dataset with spatial coordinates.
            attr_name:
                Name of the attribute to be added to the cells.
                If None, a random name will be generated.
                Default is None.
            **kwargs:
                cover_crs:
                    Whether to cover the crs of the input data.
                    If False, it assumes the input data has crs info.
                    If True, it will cover the crs of the input data by the crs of this layer.
                    Default is False.
                resampling_method:
                    The [resampling method](https://rasterio.readthedocs.io/en/stable/api/rasterio.enums.html#rasterio.enums.Resampling)
                    when re-projecting the input data.
                    Default is "nearest".
                flipud:
                    Whether to flip the input data upside down.
                    Set to True when the input data is not in the same direction as the raster layer.
                    Default is False.
        """

    def to_file(
        self,
        raster_file: str,
        attr_name: Optional[str] = None,
        driver: str = "GTiff",
    ) -> None:
        """
        Writes a raster layer to a file.

        Parameters:
            raster_file:
                The path to the raster file to write to.
            attr_name:
                The name of the attribute to write to the raster.
                If None, all attributes are written. Default is None.
            driver:
                The GDAL driver to use for writing the raster file.
                Default is 'GTiff'.
                See GDAL docs at https://gdal.org/drivers/raster/index.html.
        """

    def get_raster(self, attr_name: Optional[str] = None) -> np.ndarray:
        """Obtaining the Raster layer by attribute.

        Parameters:
            attr_name:
                The attribute to retrieve.
                If None (by default), retrieve all attributes as a 3D array.

        Returns:
            A 3D array of attribute.
        """

    @overload
    def get_neighborhood(
        self,
        pos: Coordinate,
        moore: bool,
        include_center: bool = False,
        radius: int = 1,
        annular: bool = False,
        return_mask: bool = True,
    ) -> np.ndarray:
        ...

    def get_neighborhood(
        self,
        pos: Coordinate,
        moore: bool,
        include_center: bool = False,
        radius: int = 1,
        annular: bool = False,
        return_mask: bool = False,
    ) -> ActorsList[PatchCell] | np.ndarray:
        """Get neighbors of the given position.
        This is a protocol method.
        """


class BaseNature(CompositeModule, _PatchModuleProtocol):
    """The Base Nature Module.
    Note:
        Look at [this tutorial](../tutorial/beginner/organize_model_structure.ipynb) to understand the model structure.
        This is NOT a raster layer, but can be seen as a container of different raster layers.
        Users can create new raster layer (i.e., `PatchModule`) by `new` method.
        By default, an initialized ABSESpy model will init an instance of this `BaseNature` as `nature` module.

    Attributes:
        major_layer:
            The major layer of nature module. By default, it's the first layer that user created.
        total_bounds:
            The spatial scope of the model's concern. By default, uses the major layer of this model.
    """

    def __init__(
        self, model: MainModel[Any, Any], name: str = "nature"
    ) -> None:
        CompositeModule.__init__(self, model, name=name)
        self._major_layer: Optional[PatchModule] = None
        self._modules: _PatchModuleFactory = _PatchModuleFactory(self)

    def __repr__(self) -> str:
        major_layer = (
            self.major_layer.name
            if self.major_layer is not None
            else "No major"
        )
        flag = "open" if self.opening else "closed"
        return f"<nature ({major_layer}): {flag}>"

    def __getattribute__(self, name: str) -> PatchModule:
        """Get the submodule by name."""
        if name.startswith("_"):
            return super().__getattribute__(name)
        if hasattr(_PatchModuleProtocol, name):
            return getattr(self.major_layer, name)
        return super().__getattribute__(name)

    def __getattr__(self, name: str) -> PatchModule:
        if name in self._modules:
            return self._modules[name]
        raise AttributeError(f"Unknown module {name}")

    @property
    def major_layer(self) -> Optional[PatchModule]:
        """The major layer of nature module.
        By default, it's the first created layer.
        """
        return self._major_layer

    @major_layer.setter
    def major_layer(self, layer: PatchModule) -> None:
        if not isinstance(layer, PatchModule):
            raise TypeError(f"{layer} is not PatchModule.")
        self._major_layer = layer

    @property
    def total_bounds(self) -> np.ndarray:
        """Total bounds. The spatial scope of the model's concern.
        If None (by default), uses the major layer of this model.
        Usually, the major layer is the first layer sub-module you created.
        """
        if self.major_layer is None:
            raise ValueError(f"No major layer in {self.modules}.")
        return self.major_layer.total_bounds

    @property
    def crs(self) -> pyproj.CRS:
        """Geo CRS."""
        return (
            pyproj.CRS(CRS)
            if self.major_layer is None
            else self.major_layer.crs
        )

    def create_module(
        self,
        module_cls: Optional[Type[PatchModule]] = None,
        how: Optional[HowCreation] = None,
        major_layer: bool = False,
        **kwargs: Any,
    ) -> PatchModule:
        """Creates a submodule of the raster layer.

        Parameters:
            module_cls:
                The custom module class.
            how:
                Class method to call when creating the new sub-module (raster layer).
                So far, there are three options:
                    `from_resolution`: by selecting shape and resolution.
                    `from_file`: by input of a geo-tiff dataset.
                    `copy_layer`: by copying shape, resolution, bounds, crs, and coordinates of an existing submodule.
                if None (by default), just simply create a sub-module without any custom methods (i.e., use the base class `PatchModule`).
            **kwargs:
                Any other arg passed to the creation method.
                See corresponding method of your how option from `PatchModule` class methods.

        Returns:
            the created new module.
        """
        if self.modules.is_empty:
            major_layer = True
        module = self.modules.new(how=how, module_class=module_cls, **kwargs)
        # 如果是第一个创建的模块,则将其作为主要的图层
        if major_layer:
            self.major_layer = module
        setattr(self, module.name, module)
        return module



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""This file is for multiple-run experiment.
"""
from __future__ import annotations

import copy
import itertools
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from numbers import Number
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

import numpy as np
import pandas as pd
from hydra import compose, initialize
from hydra.core.global_hydra import GlobalHydra
from hydra.core.hydra_config import HydraConfig
from omegaconf import DictConfig, OmegaConf
from tqdm.auto import tqdm

from abses.main import BaseHuman, BaseNature, MainModel, SubSystemType

Configurations: TypeAlias = DictConfig | str | Dict[str, Any]


def convert_to_python_type(value: Any) -> Any:
    """Convert numpy types to python native types.
    This function is mainly for `OmegaConfig` module to handle the parameters.
    """
    # If generic one value.
    if isinstance(value, np.generic):
        return value.item()
    # If array
    elif isinstance(value, np.ndarray):
        # Optionally convert arrays to list if necessary
        return value.tolist()
    return value


def relative_path_from_to(from_path: Path, to_path: Path) -> Path:
    """Calculate the relative path from one path to another."""
    # 将两个路径都解析为绝对路径
    from_path = from_path.resolve()
    to_path = to_path.resolve()

    # 获取两个路径的公共父目录
    common_ancestor = Path(from_path.anchor)
    # 从根目录向下寻找公共路径
    for part in from_path.parts:
        if (
            to_path.parts[: from_path.parts.index(part) + 1]
            == from_path.parts[: from_path.parts.index(part) + 1]
        ):
            common_ancestor = Path(common_ancestor) / part
        else:
            break

    # 计算从起始路径到公共父目录的距离
    relative_from = Path(
        *[".."] * (len(from_path.relative_to(common_ancestor).parts))
    )
    # 计算从公共父目录到目标路径的距离
    relative_to = to_path.relative_to(common_ancestor)

    # 组合两个部分得到最终的相对路径
    return relative_from / relative_to


class Experiment:
    """Repeated Experiment."""

    _instance = None
    folder: Path = Path(os.getcwd())
    hydra_config: DictConfig = DictConfig({})
    exp_config: DictConfig = DictConfig({})
    name: str = "ABSESpyExp"
    final_vars: List[Dict[str, Any]] = []
    model_vars: List[pd.DataFrame] = []

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        model_cls: Type[MainModel],
        nature_cls: Optional[Type[BaseNature]] = None,
        human_cls: Optional[Type[BaseHuman]] = None,
    ):
        self._n_runs = 0
        self._types: Dict[SubSystemType, Type] = {
            "model": model_cls,
            "nature": nature_cls or BaseNature,
            "human": human_cls or BaseHuman,
        }
        self._overrides: Dict[str, Any] = {}
        self._job_id: int = 0

    @classmethod
    def _update_config(
        cls,
        cfg: DictConfig,
        repeats: Optional[int] = None,
        number_process: Optional[int] = None,
    ) -> Tuple[int, int]:
        """Loading the configuration of this exp."""
        exp_config = cfg.get("exp", DictConfig({}))
        cls.exp_config = exp_config
        if repeats is None:
            repeats = exp_config.get("repeats", 1)
        if number_process is None:
            number_process = exp_config.get("num_process", 1)

        cls.name = exp_config.get("name", "ABSESpyExp")
        if cls.is_hydra_job():
            cls.hydra_config: DictConfig = HydraConfig.get()
            cls.folder = Path(HydraConfig.get().run.dir)
        return repeats, number_process

    @property
    def model(self) -> Optional[Type[MainModel]]:
        """Model class."""
        model = self._types["model"]
        if not issubclass(model, MainModel):
            raise TypeError(f"Type {type(model)} is invalid.")
        return model

    @property
    def outpath(self) -> Path:
        """Output dir path."""
        if self.is_hydra_job():
            return Path(self.hydra_config.runtime.output_dir)
        return self.folder

    @property
    def overrides(self) -> Dict[str, Any]:
        """Overrides"""
        if not self.is_hydra_job():
            return self._overrides
        overrides_dict = {}
        for item in self.hydra_config.overrides.task:
            if "=" in item:  # 确保是键值对形式的覆盖
                key, value = item.split("=", 1)
                overrides_dict[key] = value
        return overrides_dict

    @overrides.setter
    def overrides(self, current_overrides):
        """Set the overrides."""
        if not isinstance(current_overrides, dict):
            raise TypeError("current_overrides must be a dictionary.")
        self._overrides = current_overrides

    @property
    def job_id(self) -> int:
        """Job id.
        Each job means a combination of the configuration.
        If the experiment is running in Hydra, it will return the hydra's job id.
        """
        if self.is_hydra_job():
            return self.hydra_config.job.get("id", 0)
        return self._job_id

    @staticmethod
    def is_hydra_job() -> bool:
        """Returns True if the experiment is running in Hydra."""
        return GlobalHydra().is_initialized()

    def _is_config_path(self, cfg: Configurations) -> bool:
        if isinstance(cfg, str):
            return True
        if isinstance(cfg, (DictConfig, dict)):
            return False
        raise TypeError(f"Unknown config type {type(cfg)}.")

    def _overriding_config(
        self,
        cfg: DictConfig | Dict[str, Any],
        overrides: Optional[Dict[str, str | Iterable[Number]]] = None,
    ) -> Iterator[DictConfig]:
        """Parse the config."""
        if overrides is None:
            return iter([cfg])
        if isinstance(cfg, dict):
            cfg = DictConfig(cfg)
        keys, values = zip(*overrides.items())
        values = tuple(convert_to_python_type(val) for val in values)
        combinations = itertools.product(*values)
        for comb in combinations:
            cfg_copy = copy.deepcopy(cfg)
            current_overrides = dict(zip(keys, comb))
            for key, val in current_overrides.items():
                OmegaConf.update(cfg_copy, key=key, value=val, merge=True)
            yield cfg_copy, current_overrides

    def _load_hydra_cfg(
        self,
        cfg_path: Path,
        overrides: Optional[Dict[str, str | Iterable[Number]]] = None,
    ) -> Optional[DictConfig]:
        """Initialize Hydra with overrides."""
        if self.is_hydra_job():
            return HydraConfig.get().cfg
        with initialize(version_base=None, config_path=str(cfg_path.parent)):
            cfg = compose(config_name=cfg_path.stem, overrides=overrides)
        return cfg

    def _get_logging_mode(self, repeat_id: Optional[int] = None) -> str | bool:
        log_mode = self.exp_config.get("logging", "once")
        if log_mode == "once":
            if repeat_id == 1:
                logging: bool | str = self.name
            else:
                return False
        elif bool(log_mode):
            logging = f"{self.name}_{repeat_id}"
        else:
            logging = False
        return logging

    def _update_log_config(
        self, config, repeat_id: Optional[int] = None
    ) -> bool:
        """Update the log configuration."""
        if isinstance(config, dict):
            config = DictConfig(config)
        OmegaConf.set_struct(config, False)
        log_name = self._get_logging_mode(repeat_id=repeat_id)
        if not log_name:
            config["log"] = False
            return config
        logging_cfg = OmegaConf.create({"log": {"name": log_name}})
        config = OmegaConf.merge(config, logging_cfg)
        return config

    def run(
        self, cfg: DictConfig, repeat_id: int, outpath: Optional[Path] = None
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        """运行模型一次"""
        cfg = self._update_log_config(cfg, repeat_id)
        # 获取日志
        model = self.model(
            parameters=cfg,
            run_id=repeat_id,
            outpath=outpath,
            experiment=self,
            nature_class=self._types["nature"],
            human_class=self._types["human"],
        )
        model.run_model()
        if model.datacollector.model_reporters:
            df = model.datacollector.get_model_vars_dataframe()
        else:
            df = pd.DataFrame()
        final_report = model.datacollector.get_final_vars_report(model)
        return final_report, df

    def _update_result(
        self,
        repeat_id: int,
        reports: Optional[Dict[str, Any]] = None,
        model_df: Optional[pd.DataFrame] = None,
    ) -> None:
        """Updating in each run."""
        if reports is None:
            reports = {}
        reports.update(
            {
                "job_id": self.job_id,
                "repeat_id": repeat_id,
            }
        )
        reports |= self.overrides
        self.final_vars.append(reports)
        if model_df is None:
            return
        model_df = model_df.reset_index()
        model_df = model_df.rename({"index": "tick"}, axis=1)
        model_df.insert(0, "repeat_id", repeat_id)
        model_df.insert(0, "job_id", self.job_id)
        self.model_vars.append(model_df)

    def _batch_run_multi_processes(
        self,
        cfg: DictConfig,
        repeats,
        number_process: Optional[int] = 1,
        display_progress: bool = True,
    ):
        # Multiple processes
        with tqdm(total=repeats, disable=not display_progress) as pbar:
            with ProcessPoolExecutor(max_workers=number_process) as executor:
                # 提交所有任务
                futures = [
                    executor.submit(self.run, cfg, repeat_id, self.outpath)
                    for repeat_id in range(1, repeats + 1)
                ]
                # 使用as_completed等待任务完成
                for i, future in enumerate(as_completed(futures)):
                    # 每完成一个任务，更新进度条和运行计数
                    result, model_df = future.result()
                    pbar.update()
                    self._update_result(
                        reports=result, model_df=model_df, repeat_id=i + 1
                    )

    def _batch_run_repeats(self, cfg, repeats, display_progress) -> None:
        for repeat in tqdm(range(repeats), disable=not display_progress):
            result, model_df = self.run(
                cfg, repeat_id=repeat + 1, outpath=self.outpath
            )
            self._update_result(
                reports=result, model_df=model_df, repeat_id=repeat + 1
            )

    def _parse_path(self, relative_path: str) -> Path:
        """Parse the path of the configuration file.
        Convert the relpath of current work space to the relpath of this script.
        """
        # 目标绝对路径
        abs_config_file_path = (Path(os.getcwd()) / relative_path).resolve()
        if not abs_config_file_path.is_file():
            raise FileNotFoundError(f"File {abs_config_file_path} not found.")
        # 返回相对于本脚本的路径
        current_file_path = Path(__file__).parent.resolve()
        return relative_path_from_to(current_file_path, abs_config_file_path)

    def batch_run(
        self,
        cfg: Configurations,
        repeats: Optional[int] = None,
        parallels: Optional[int] = None,
        display_progress: bool = True,
        overrides: Optional[Dict[str, str | Iterable[Number]]] = None,
    ) -> None:
        """Run the experiment multiple times.

        Parameters:
            cfg:
                The configuration of the experiment.
                It can be either a string of the config file path,
                or a dictionary of the config.
                For an example:
                `cfg='config.yaml'` refers to the config file `config.yaml` saved in the current workspace.
                `cfg={'exp': {'name': 'test', 'repeats': 2}}` refers to the config dictionary.
            repeats:
                The number of repeats for the experiment.
                If not specified, it will use the default value 1 (No repeats).
            parallels:
                The number of processes running in parallel.
                If not specified, it will use the default value 1 (No parallel).
            display_progress:
                Whether to display the progress bar, default True.
            overrides:
                The dictionary of overrides for the experiment.
                If specified, the experiment will sweep all the possible values for the parameter.
                For examples:
                override = {model.key: ["cate1", "cate2"]}
                override = {nature.key: np.arange(10, 2)}
                The first override will lead to two different runs:
                - model.key = cate1
                - model.key = cate2
                The second override will lead to a series runs:
                - model.nature.key = 0.0
                - model.nature.key = 2.0
                - model.nature.key = 4.0
                - model.nature.key = 6.0
                - model.nature.key = 8.0

        Example:
            ```Python
            # initialize the experiment.
            exp = Experiment(MainModel)
            # Loading the configuration file `config.yaml`.
            exp.batch_run('config.yaml')
            ```

            ```Python
            # A different way for initializing.
            exp = Experiment.new(MainModel)
            cfg = {'time': {'end': 25}}

            # Nine runs with different ending ticks.
            exp.batch_run(cfg=cfg, overrides={'time.end': range(10, 100, 10)})
            ```
        """
        # 如果配置是路径，则利用 Hydra API先清洗配置
        if self._is_config_path(cfg):
            config_path = self._parse_path(cast(str, cfg))
            cfg = self._load_hydra_cfg(config_path)
        # 加载配置
        repeats, parallels = self._update_config(cfg, repeats, parallels)
        # 如果没有指定覆写，则直接运行程序
        if not overrides:
            if parallels == 1 or repeats == 1:
                self._batch_run_repeats(cfg, repeats, display_progress)
            else:
                self._batch_run_multi_processes(
                    cfg,
                    repeats,
                    parallels,
                    display_progress,
                )
        # 否则，对每一个复写的配置进行重复运行
        for config, current_overrides in self._overriding_config(
            cfg, overrides
        ):
            self.overrides = current_overrides
            self.batch_run(
                cfg=config,
                repeats=repeats,
                parallels=parallels,
                display_progress=display_progress,
            )
            self._job_id += 1
        # finally, clean up the overrides now.
        self.overrides = {}

    @classmethod
    def summary(cls, save: bool = False, **kwargs) -> None:
        """Ending the experiment."""
        df = pd.DataFrame(cls.final_vars)
        if save:
            df.to_csv(cls.folder / "summary.csv", index=False, **kwargs)
        return df

    @classmethod
    def clean(cls, new_exp: bool = False) -> None:
        """Clean the results.

        Parameters:
            new_exp:
                Whether to create a new experiment.
                If True, it will delete all the current settings.
        """
        cls.final_vars = []
        cls.model_vars = []
        if new_exp:
            cls._instance = None
            cls.folder = Path(os.getcwd())
            cls.hydra_config = DictConfig({})
            cls.name = "ABSESpyExp"

    @classmethod
    def new(cls, model_cls: Type[MainModel]) -> Experiment:
        """Create a new experiment for the singleton class `Experiment`.
        This method will delete all currently available exp results and settings.
        Then, it initialize a new instance of experiment.

        Parameters:
            model_cls:
                Using which model class to initialize the experiment.

        Raises:
            TypeError:
                If the model class `model_cls` is not a valid `ABSESpy` model.

        Returns:
            An experiment.
        """
        cls.clean(new_exp=True)
        return cls(model_cls=model_cls)

    @classmethod
    def get_model_vars_dataframe(cls) -> pd.DataFrame:
        """Aggregation of model vars dataframe."""
        if cls.model_vars:
            return pd.concat(cls.model_vars, axis=0)
        raise ValueError("No model vars found.")



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""PatchModule class
"""
from __future__ import annotations

import copy
import functools
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

import geopandas as gpd
import numpy as np
import pyproj
import rasterio
import rioxarray
import xarray as xr
from geocube.api.core import make_geocube
from loguru import logger
from mesa.space import Coordinate
from mesa_geo.raster_layers import RasterBase
from rasterio import mask
from rasterio.enums import Resampling
from rasterio.warp import calculate_default_transform, transform_bounds
from shapely import Geometry

from abses._bases.errors import ABSESpyError
from abses._bases.modules import Module, _ModuleFactory
from abses.cells import PatchCell
from abses.random import ListRandom
from abses.sequences import ActorsList
from abses.tools.func import get_buffer, set_null_values
from abses.viz.viz_nature import _VizNature

if TYPE_CHECKING:
    from abses.main import MainModel

CRS = "epsg:4326"
CellFilter: TypeAlias = Optional[str | np.ndarray | xr.DataArray | Geometry]
Raster: TypeAlias = Union[
    np.ndarray,
    xr.DataArray,
    xr.Dataset,
]


class _PatchModuleFactory(_ModuleFactory):
    def __init__(self, father) -> None:
        super().__init__(father)
        self.default_cls = PatchModule

    def __getitem__(self, name: str) -> PatchModule:
        return self.modules[name]

    def from_resolution(
        self,
        model: MainModel[Any, Any],
        name: Optional[str] = None,
        shape: Coordinate = (10, 10),
        crs: Optional[pyproj.CRS | str] = CRS,
        resolution: Union[int, float] = 1,
        module_cls: Optional[type[PatchModule]] = None,
        cell_cls: type[PatchCell] = PatchCell,
    ) -> PatchModule:
        """Create a layer from resolution.

        Parameters:
            model:
                ABSESpy Model that the new module belongs.
            name:
                Name of the new module.
                If None (by default), using lowercase of the '__class__.__name__'.
                E.g., class Module -> module.
            shape:
                Array shape (height, width) of the new module.
                For example, `shape=(3, 5)` means the new module stores 15 cells.
            crs:
                Coordinate Reference Systems.
                If passing a string object, should be able to parsed by `pyproj`.
                By default, we use CRS = "epsg:4326".
            resolution:
                Spatial Resolution when creating the coordinates.
                By default 1, it means shape (3, 5) will generate coordinates:
                {y: [0, 1, 2], x: [0, 1, 2, 3, 4]}.
                Similar, when using resolution=0.1,
                it will be {y: [.0, .1, .2], x: [.0, .1, .2, .3, .4]}.
            cell_cls:
                Class type of `PatchCell` to create.

        Returns:
            A new instance of self ("PatchModule").
        """
        to_create = cast(PatchModule, self._check_cls(module_cls=module_cls))
        height, width = shape
        total_bounds = [0, 0, width * resolution, height * resolution]
        return to_create(
            model,
            name=name,
            width=width,
            height=height,
            crs=crs,
            total_bounds=total_bounds,
            cell_cls=cell_cls,
        )

    def copy_layer(
        self,
        model: MainModel[Any, Any],
        layer: PatchModule,
        name: Optional[str] = None,
        module_cls: Optional[Type[PatchModule]] = None,
        cell_cls: Type[PatchCell] = PatchCell,
    ) -> PatchModule:
        """Copy an existing layer to create a new layer.

        Parameters:
            model:
                ABSESpy Model that the new module belongs.
            layer:
                Another layer to copy.
                These attributes will be copied:
                including the coordinates, the crs, and the shape.
            name:
                Name of the new module.
                If None (by default), using lowercase of the '__class__.__name__'.
                E.g., class Module -> module.
            cell_cls:
                Class type of `PatchCell` to create.

        Returns:
            A new instance of self ("PatchModule").
        """
        if not isinstance(layer, PatchModule):
            raise TypeError(f"{layer} is not a valid PatchModule.")
        to_create = cast(PatchModule, self._check_cls(module_cls=module_cls))
        return to_create(
            model=model,
            name=name,
            width=layer.width,
            height=layer.height,
            crs=layer.crs,
            total_bounds=layer.total_bounds,
            cell_cls=cell_cls,
        )

    def from_xarray(
        self,
        xda: xr.DataArray,
        model: MainModel[Any, Any],
        module_cls: Optional[Type[PatchModule]] = None,
        name: str | None = None,
        attr_name: str | None = None,
        apply_raster: bool = False,
        masked: bool = True,
        cell_cls: type[PatchCell] = PatchCell,
        **kwargs,
    ) -> PatchModule:
        """Create a new module instance from `xarray.DataArray` data."""
        # 如果 y 轴是从小到大的，反转它
        if xda.y[0].item() < xda.y[-1].item():
            xda.data = np.flipud(xda.data)
        # 创建模块
        to_create = cast(PatchModule, self._check_cls(module_cls=module_cls))
        module: PatchModule = to_create(
            model=model,
            name=name,
            width=xda.rio.width,
            height=xda.rio.height,
            crs=xda.rio.crs,
            total_bounds=xda.rio.bounds(),
            cell_cls=cell_cls,
        )
        if masked:
            module.mask = xda.notnull().to_numpy()
        if apply_raster:
            module.apply_raster(xda.to_numpy(), attr_name=attr_name, **kwargs)
        return module

    def from_vector(
        self,
        vector_file: str | Path | gpd.GeoDataFrame,
        resolution: Tuple[float, float] | float,
        model: MainModel[Any, Any],
        module_cls: Optional[Type[PatchModule]] = None,
        name: str | None = None,
        attr_name: Optional[str] = None,
        apply_raster: bool = False,
        masked: bool = True,
        cell_cls: type[PatchCell] = PatchCell,
    ) -> PatchModule:
        """Create a layer module from a shape file."""
        if isinstance(vector_file, (str, Path)):
            gdf = gpd.read_file(vector_file)
        elif isinstance(vector_file, gpd.GeoDataFrame):
            gdf = vector_file
        else:
            raise TypeError(f"Unsupported vector {type(vector_file)}.")
        if attr_name is None:
            gdf, attr_name = gdf.reset_index(), "index"
        if isinstance(resolution, float):
            resolution = (resolution, resolution)
        xda = make_geocube(
            gdf, measurements=[attr_name], resolution=resolution
        )[attr_name]
        return self.from_xarray(
            xda=xda,
            model=model,
            module_cls=module_cls,
            name=name,
            attr_name=attr_name,
            apply_raster=apply_raster,
            masked=masked,
            cell_cls=cell_cls,
        )

    def from_file(
        self,
        raster_file: str,
        model: MainModel[Any, Any],
        cell_cls: type[PatchCell] = PatchCell,
        module_cls: Optional[Type[PatchModule]] = None,
        name: str | None = None,
        attr_name: str | None = None,
        apply_raster: bool = False,
        band: int = 1,
        masked: bool = True,
        **kwargs: Any,
    ) -> PatchModule:
        """Create a raster layer module from a file.

        Parameters:
            raster_file:
                File path of a geo-tiff dataset.
            model:
                ABSESpy Model that the new module belongs.
            attr_name:
                Assign a attribute name to the loaded raster data.
            name:
                Name of the new module.
                If None (by default), using lowercase of the '__class__.__name__'.
                E.g., class Module -> module.
            cell_cls:
                Class type of `PatchCell` to create.

        """
        xda = rioxarray.open_rasterio(raster_file, masked=masked, **kwargs)
        xda = xda.sel(band=band)
        return self.from_xarray(
            xda=xda,
            model=model,
            module_cls=module_cls,
            name=name,
            attr_name=attr_name,
            apply_raster=apply_raster,
            masked=masked,
            cell_cls=cell_cls,
        )


class PatchModule(Module, RasterBase):
    """
    The spatial sub-module base class.
    Inherit from this class to create a submodule.
    [This tutorial](../tutorial/beginner/organize_model_structure.ipynb) shows the model structure.
    This is also a raster layer, inherited from the 'mesa-geo.RasterLayer' class.
    ABSESpy extends this class, so it can:
    1. place agents (by `_CellAgentsContainer` class.)
    2. work with `xarray`, `rasterio` packages for better data I/O workflow.

    Attributes:
        cell_properties:
            The accessible attributes of cells stored in this layer.
            When a `PatchCell`'s method is decorated by `raster_attribute`,
            it should be appeared here as a property attribute.
        attributes:
            All accessible attributes from this layer,
            including cell_properties.
        shape2d:
            Raster shape in 2D (heigh, width).
        shape3d:
            Raster shape in 3D (1, heigh, width),
            this is for compatibility with `mg.RasterLayer` and `rasterio`.
        array_cells:
            Array type of the `PatchCell` stored in this module.
        coords:
            Coordinate system of the raster data.
            This is useful when working with `xarray.DataArray`.
        random:
            A random proxy by calling the cells as an `ActorsList`.
    """

    def __init__(
        self,
        model: MainModel[Any, Any],
        name: Optional[str] = None,
        cell_cls: Type[PatchCell] = PatchCell,
        **kwargs: Any,
    ):
        """This method copied some of the `mesa-geo.RasterLayer`'s methods."""
        Module.__init__(self, model, name=name)
        RasterBase.__init__(self, **kwargs)
        self.cell_cls = cell_cls
        logger.info("Initializing a new Model Layer...")
        logger.info(f"Using rioxarray version: {rioxarray.__version__}")

        func = np.vectorize(lambda row, col: cell_cls(self, (row, col)))
        self._cells: np.ndarray = np.fromfunction(
            func, shape=(self.height, self.width), dtype=object
        )
        self._attributes: Set[str] = set()
        self.mask: np.ndarray = np.ones(self.shape2d)

    @functools.cached_property
    def cells(self) -> ActorsList[PatchCell]:
        """The cells stored in this layer."""
        return ActorsList(self.model, self.array_cells[self.mask])

    @property
    def mask(self) -> np.ndarray:
        """Where is not accessible."""
        return self._mask

    @mask.setter
    def mask(self, array: np.ndarray) -> None:
        """Setting mask."""
        if array.shape != self.shape2d:
            raise ABSESpyError(
                f"Shape mismatching, setting mask {array.shape}."
                f"but the module is expecting shape {self.shape2d}."
            )
        self._mask = array.astype(bool)

    def __repr__(self):
        return f"<{self.name}{self.shape2d}: {len(self.attributes)} vars>"

    def __getitem__(
        self,
        index: int | Sequence[Coordinate] | tuple[int | slice, int | slice],
    ) -> PatchCell | list[PatchCell]:
        """
        Access contents from the grid.
        """
        return self.array_cells.__getitem__(index)

    def __iter__(self) -> Iterator[PatchCell]:
        """
        Create an iterator that chains the rows of the cells together
        as if it is one list
        """
        for row in self.array_cells:
            yield from row

    @property
    def cell_properties(self) -> set[str]:
        """The accessible attributes of cells stored in this layer.
        All `PatchCell` methods decorated by `raster_attribute` should be appeared here.
        """
        return self.cell_cls.__attribute_properties__()

    @functools.cached_property
    def xda(self) -> xr.DataArray:
        """Get the xarray raster layer with spatial coordinates."""
        xda = xr.DataArray(data=self.mask, coords=self.coords)
        xda = xda.rio.write_crs(self.crs)
        xda = xda.rio.set_spatial_dims("x", "y")
        xda = xda.rio.write_transform(self.transform)
        return xda.rio.write_coordinate_system()

    @functools.cached_property
    def plot(self) -> _VizNature:
        """Plotting"""
        return _VizNature(self)

    @property
    def attributes(self) -> set[str]:
        """All accessible attributes from this layer."""
        return self._attributes | self.cell_properties

    @property
    def shape2d(self) -> Coordinate:
        """Raster shape in 2D (height, width).
        This is useful when working with 2d `numpy.array`.
        """
        return self.height, self.width

    @property
    def shape3d(self) -> Coordinate:
        """Raster shape in 3D (1, heigh, width).
        This is useful when working with `rasterio` band.
        """
        return 1, self.height, self.width

    @functools.cached_property
    def array_cells(self) -> np.ndarray:
        """Array type of the `PatchCell` stored in this module."""
        return self._cells

    @property
    def coords(self) -> Coordinate:
        """Coordinate system of the raster data.

        This is useful when working with `xarray.DataArray`.
        """
        nrows, ncols = self.shape2d
        coords_x = np.arange(ncols) * self.transform[0] + self.transform[2]
        coords_y = np.arange(nrows) * self.transform[4] + self.transform[5]
        return {
            "y": coords_y,
            "x": coords_x,
        }

    def transform_coord(self, row: int, col: int) -> Coordinate:
        """Transforming the row, col to the real-world coordinate."""
        if self.out_of_bounds(pos=(row, col)):
            raise IndexError(f"Out of bounds: {row, col}")
        return self.transform * (col, row)

    def to_crs(self, crs, inplace=False) -> Optional[PatchModule]:
        """Converting the raster data to a another CRS."""
        super()._to_crs_check(crs)
        layer = self if inplace else copy.copy(self)

        src_crs = rasterio.crs.CRS.from_user_input(layer.crs)
        dst_crs = rasterio.crs.CRS.from_user_input(crs)
        if not layer.crs.is_exact_same(crs):
            transform, _, _ = calculate_default_transform(
                src_crs,
                dst_crs,
                self.width,
                self.height,
                *layer.total_bounds,
            )
            layer.total_bounds = [
                *transform_bounds(src_crs, dst_crs, *layer.total_bounds)
            ]
            layer.crs = crs
            layer.transform = transform

        return None if inplace else layer

    def _attr_or_array(
        self, data: None | str | np.ndarray | xr.DataArray
    ) -> np.ndarray:
        """Determine the incoming data type and turn it into a reasonable array."""
        if data is None:
            return np.ones(self.shape2d)
        if isinstance(data, xr.DataArray):
            data = data.to_numpy()
        if isinstance(data, np.ndarray):
            if data.shape == self.shape2d:
                return data
            raise ABSESpyError(
                f"Shape mismatch: {data.shape} [input] != {self.shape2d} [expected]."
            )
        if isinstance(data, str) and data in self.attributes:
            return self.get_raster(data)
        raise TypeError("Invalid data type or shape.")

    def dynamic_var(
        self,
        attr_name: str,
        dtype: Literal["numpy", "xarray", "rasterio"] = "numpy",
    ) -> np.ndarray:
        """Update and get dynamic variable.

        Parameters:
            attr_name:
                The dynamic variable to retrieve.

        Returns:
            2D numpy.ndarray data of the variable.
        """
        # 获取动态变量，及其附加属性
        array = super().dynamic_var(attr_name)
        assert isinstance(array, (np.ndarray, xr.DataArray, xr.Dataset))
        kwargs = super().dynamic_variables[attr_name].attrs
        # 将矩阵转换为三维，并更新空间数据
        self.apply_raster(array, attr_name=attr_name, **kwargs)
        if dtype == "numpy":
            return self.get_raster(attr_name, update=False)
        if dtype == "xarray":
            return self.get_xarray(attr_name, update=False)
        if dtype == "rasterio":
            return self.get_rasterio(attr_name, update=False)
        raise ValueError(f"Unknown expected dtype {dtype}.")

    def get_rasterio(
        self,
        attr_name: Optional[str] = None,
        update: bool = True,
    ) -> rasterio.MemoryFile:
        """Gets the Rasterio raster layer corresponding to the attribute. Save to a temporary rasterio memory file.

        Parameters:
            attr_name:
                The attribute name for creating the rasterio file.

        Returns:
            The rasterio tmp memory file of raster.
        """
        if attr_name is None:
            data = np.ones(self.shape2d)
        else:
            data = self.get_raster(attr_name=attr_name, update=update)
        # 如果获取到的是2维，重整为3维
        if len(data.shape) != 3:
            data = data.reshape(self.shape3d)
        with rasterio.MemoryFile() as mem_file:
            with mem_file.open(
                driver="GTiff",
                height=data.shape[1],
                width=data.shape[2],
                count=data.shape[0],  # number of bands
                dtype=str(data.dtype),
                crs=self.crs,
                transform=self.transform,
            ) as dataset:
                dataset.write(data)
            # Open the dataset again for reading and return
            return mem_file.open()

    def get_xarray(
        self,
        attr_name: Optional[str] = None,
        update: bool = True,
    ) -> xr.DataArray:
        """Get the xarray raster layer with spatial coordinates.

        Parameters:
            attr_name:
                The attribute to retrieve. If None (by default),
                return all available attributes (3D DataArray).
                Otherwise, 2D DataArray of the chosen attribute.

        Returns:
            Xarray.DataArray data with spatial coordinates of the chosen attribute.
        """
        data = self.get_raster(attr_name=attr_name, update=update)
        if attr_name:
            name = attr_name
            data = data.reshape(self.shape2d)
            coords = self.coords
        else:
            coords = {"variable": list(self.attributes)}
            coords |= self.coords
            name = self.name
        return xr.DataArray(
            data=data,
            name=name,
            coords=coords,
        ).rio.write_crs(self.crs)

    @property
    def random(self) -> ListRandom:
        """Randomly"""
        return self.cells.random

    def _select_by_geometry(
        self,
        geometry: Geometry,
        refer_layer: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> np.ndarray:
        """Gets all the cells that intersect the given geometry.

        Parameters:
            geometry:
                Shapely Geometry to search intersected cells.
            refer_layer:
                The attribute name to refer when filtering cells.
            **kwargs:
                Args pass to the function `rasterasterio.mask.mask`. It influence how to build the mask for filtering cells. Please refer [this doc](https://rasterasterio.readthedocs.io/en/latest/api/rasterasterio.mask.html) for details.

        Raises:
            ABSESpyError:
                If no available attribute exists, or the assigned refer layer is not available in the attributes.

        Returns:
            A list of PatchCell.
        """
        if refer_layer is not None and refer_layer not in self.attributes:
            raise ABSESpyError(
                f"The refer layer {refer_layer} is not available in the attributes"
            )
        data = self.get_rasterio(attr_name=refer_layer)
        out_image, _ = mask.mask(data, [geometry], **kwargs)
        return out_image.reshape(self.shape2d)

    def select(
        self,
        where: Optional[CellFilter] = None,
    ) -> ActorsList[PatchCell]:
        """Select cells from this layer.
        Also has a shortcut alias for this method: `.sel`.

        Parameters:
            where:
                The condition to select cells.
                If None (by default), select all cells.
                If a string, select cells by the attribute name.
                If a numpy.ndarray, select cells by the mask array.
                If a Shapely Geometry, select cells by the intersection with the geometry.

        Raises:
            TypeError:
                If the input type is not supported.

        Returns:
            An `ActorsList` with all selected cells stored.
        """
        if isinstance(where, Geometry):
            mask_ = self._select_by_geometry(geometry=where)
        elif (
            isinstance(where, (np.ndarray, str, xr.DataArray)) or where is None
        ):
            mask_ = self._attr_or_array(where).reshape(self.shape2d)
        else:
            raise TypeError(
                f"{type(where)} is not supported for selecting cells."
            )
        mask_ = np.nan_to_num(mask_, nan=0.0).astype(bool) & self.mask
        return ActorsList(self.model, self.array_cells[mask_])

    sel = select

    def apply(
        self, ufunc: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> np.ndarray:
        """Apply a function to array cells.

        Parameters:
            ufunc:
                A function to apply.
            *args:
                Positional arguments to pass to the function.
            **kwargs:
                Keyword arguments to pass to the function.

        Returns:
            The result of the function applied to the array cells.
        """
        func = functools.partial(ufunc, *args, **kwargs)
        return np.vectorize(func)(self.array_cells, *args, **kwargs)

    def coord_iter(self) -> Iterator[tuple[Coordinate, PatchCell]]:
        """
        An iterator that returns coordinates as well as cell contents.
        """
        return np.ndenumerate(self.array_cells)

    def _add_attribute(
        self,
        data: np.ndarray,
        attr_name: Optional[str] = None,
        flipud: bool = False,
        apply_mask: bool = False,
    ) -> None:
        try:
            data = data.reshape(self.shape2d)
        except ValueError as e:
            raise ValueError(
                f"Data shape does not match raster shape. "
                f"Expected {self.shape2d}, received {data.shape}."
            ) from e
        if apply_mask:
            set_null_values(data, ~self.mask)
        if attr_name is None:
            attr_name = f"attribute_{len(self.attributes)}"
        self._attributes.add(attr_name)
        if flipud:
            data = np.flipud(data)
        np.vectorize(setattr)(self.array_cells, attr_name, data)

    def _add_dataarray(
        self,
        data: xr.DataArray,
        attr_name: Optional[str] = None,
        cover_crs: bool = False,
        resampling_method: str = "nearest",
        flipud: bool = False,
    ) -> None:
        if cover_crs:
            data.rio.write_crs(self.crs, inplace=True)
        resampling = getattr(Resampling, resampling_method)
        data = data.rio.reproject_match(
            self.xda,
            resampling=resampling,
        ).to_numpy()
        self._add_attribute(data, attr_name, flipud=flipud)

    def apply_raster(
        self, data: Raster, attr_name: Optional[str], **kwargs: Any
    ) -> None:
        """Apply raster data to the cells.

        Parameters:
            data:
                np.ndarray data: 2D numpy array with shape (1, height, width).
                xr.DataArray data: xarray DataArray with spatial coordinates.
                xr.Dataset data: xarray Dataset with spatial coordinates.
            attr_name:
                Name of the attribute to be added to the cells.
                If None, a random name will be generated.
                Default is None.
            **kwargs:
                cover_crs:
                    Whether to cover the crs of the input data.
                    If False, it assumes the input data has crs info.
                    If True, it will cover the crs of the input data by the crs of this layer.
                    Default is False.
                resampling_method:
                    The [resampling method](https://rasterio.readthedocs.io/en/stable/api/rasterio.enums.html#rasterio.enums.Resampling)
                    when re-projecting the input data.
                    Default is "nearest".
                flipud:
                    Whether to flip the input data upside down.
                    Set to True when the input data is not in the same direction as the raster layer.
                    Default is False.
        """
        if isinstance(data, np.ndarray):
            self._add_attribute(data, attr_name, **kwargs)
        elif isinstance(data, xr.DataArray):
            self._add_dataarray(data, attr_name, **kwargs)
        elif isinstance(data, xr.Dataset):
            if attr_name is None:
                raise ValueError("Attribute name is required for xr.Dataset.")
            dataarray = data[attr_name]
            self._add_dataarray(dataarray, attr_name, **kwargs)

    def get_raster(
        self,
        attr_name: Optional[str] = None,
        update: bool = True,
    ) -> np.ndarray:
        """Obtaining the Raster layer by attribute.

        Parameters:
            attr_name:
                The attribute to retrieve.
                If None (by default), retrieve all attributes as a 3D array.

        Returns:
            A 3D array of attribute.
        """
        if attr_name in self.dynamic_variables and update:
            return self.dynamic_var(attr_name=attr_name).reshape(self.shape3d)
        if attr_name is not None and attr_name not in self.attributes:
            raise ValueError(
                f"Attribute {attr_name} does not exist. "
                f"Choose from {self.attributes}, or set `attr_name` to `None` to retrieve all."
            )
        if attr_name is None:
            assert bool(self.attributes), "No attribute available."
            attr_names = self.attributes
        else:
            attr_names = {attr_name}
        data = []
        for name in attr_names:
            array = np.vectorize(getattr)(self.array_cells, name)
            data.append(array)
        return np.stack(data)

    def reproject(
        self,
        xda: xr.DataArray,
        resampling: Resampling | str = "nearest",
        **kwargs,
    ) -> xr.DataArray:
        """Reproject the xarray data to the same CRS as this layer."""
        if isinstance(resampling, str):
            resampling = getattr(Resampling, resampling)
        return xda.rio.reproject_match(
            self.xda, resampling=resampling, **kwargs
        )

    @overload
    def get_neighborhood(
        self,
        pos: Coordinate,
        moore: bool,
        include_center: bool = False,
        radius: int = 1,
        annular: bool = False,
        return_mask: bool = True,
    ) -> np.ndarray:
        ...

    @functools.lru_cache(maxsize=1000)
    def get_neighborhood(
        self,
        pos: Coordinate,
        moore: bool,
        include_center: bool = False,
        radius: int = 1,
        annular: bool = False,
        return_mask: bool = False,
    ) -> ActorsList[PatchCell] | np.ndarray:
        """Getting neighboring positions of the given coordinate.

        Parameters:
            pos:
                The coordinate to get the neighborhood.
            moore:
                Whether to use Moore neighborhood.
                If False, use Von Neumann neighborhood.
            include_center:
                Whether to include the center cell.
                Default is False.
            radius:
                The radius of the neighborhood.
                Default is 1.
            annular:
                Whether to use annular (ring) neighborhood.
                Default is False.
            return_mask:
                Whether to return a mask array.
                If True, return a mask array of the neighboring locations.
                Otherwise, return a `ActorsList` of neighboring cells.
                Default is False.

        Returns:
            An `ActorsList` of neighboring cells, or a mask array.
            Where the mask array is a boolean array with the same shape as the raster layer.
            The True value indicates the cell is in the neighborhood.
        """
        mask_arr = np.zeros(self.shape2d, dtype=bool)
        mask_arr[pos[0], pos[1]] = True
        mask_arr = get_buffer(
            mask_arr, radius=radius, moor=moore, annular=annular
        )
        mask_arr[pos[0], pos[1]] = include_center
        if return_mask:
            return mask_arr
        return ActorsList(self.model, self.array_cells[mask_arr])

    def to_file(
        self,
        raster_file: str,
        attr_name: Optional[str] = None,
        driver: str = "GTiff",
    ) -> None:
        """
        Writes a raster layer to a file.

        Parameters:
            raster_file:
                The path to the raster file to write to.
            attr_name:
                The name of the attribute to write to the raster.
                If None, all attributes are written. Default is None.
            driver:
                The GDAL driver to use for writing the raster file.
                Default is 'GTiff'.
                See GDAL docs at https://gdal.org/drivers/raster/index.html.
        """

        data = self.get_raster(attr_name)
        with rasterio.open(
            raster_file,
            "w",
            driver=driver,
            width=self.width,
            height=self.height,
            count=data.shape[0],
            dtype=data.dtype,
            crs=self.crs,
            transform=self.transform,
        ) as dataset:
            dataset.write(data)

    def out_of_bounds(self, pos: Coordinate) -> bool:
        """
        Determines whether position is off the grid.

        Parameters:
            pos: Position to check.

        Returns:
            True if position is off the grid, False otherwise.
        """

        row, col = pos
        return row < 0 or row >= self.height or col < 0 or col >= self.width



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Set, Union

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from abses.actor import Actor
from abses.links import _LinkContainer

from ._bases.modules import CompositeModule, Module
from .cells import PatchCell
from .container import _AgentsContainer
from .sequences import ActorsList, Selection

Actors: TypeAlias = Union[ActorsList[Actor], Selection, Actor]
Trigger: TypeAlias = Union[str, Callable[..., Any]]
if TYPE_CHECKING:
    from abses import MainModel


class HumanModule(Module):
    """The `Human` sub-module base class.

    Note:
        Look at [this tutorial](../tutorial/beginner/organize_model_structure.ipynb) to understand the model structure.

    Attributes:
        agents:
            The agents container of this ABSESpy model.
        collections:
            Actor collections defined.
    """

    def __init__(self, model: MainModel[Any, Any], name: Optional[str] = None):
        Module.__init__(self, model, name)
        self._collections: Dict[str, Selection] = {}

    @property
    def agents(self) -> _AgentsContainer:
        """The agents container of this ABSESpy model."""
        return self.model.agents

    @property
    def collections(self) -> Set[str]:
        """Actor collections defined."""
        return set(self._collections.keys())

    def actors(self, name: Optional[str] = None) -> ActorsList[Actor]:
        """Different selections of agents"""
        if name is None:
            return self.agents.get()
        if name not in self._collections:
            raise KeyError(f"{name} is not defined.")
        selection = self._collections[name]
        return self.actors().select(selection)

    def _must_be_actor(self, actor: Actor) -> None:
        if not isinstance(actor, Actor):
            raise TypeError(
                f"Actor must be a subclass of Actor, instead of {type(actor)}."
            )

    def _must_be_cell(self, cell: PatchCell) -> None:
        if not isinstance(cell, PatchCell):
            raise TypeError(
                f"Cell must be a subclass of Cell, instead of {type(cell)}."
            )

    def define(self, name: str, selection: Selection) -> ActorsList[Actor]:
        """Define a query of actors and save it into collections.

        Parameters:
            name:
                defined name of this group of actors.
            selection:
                Selection query of `Actor`.

        Raises:
            KeyError:
                If the name is already defined.

        Returns:
            The list of actors who are satisfied the query condition.

        Example:
            ```
            # Create 5 actors to query
            model=MainModel()
            model.agents.new(Actor, 5)

            module = HumanModule(model=model)
            actors = module.define(name='first', selection='ids=0')
            >>> len(actors)
            >>> 1

            >>> module.actors('first') == actors
            >>> True
            ```
        """
        if name in self._collections:
            raise KeyError(f"{name} is already defined.")
        selected = self.actors().select(selection)
        self._collections[name] = selection
        return selected


class BaseHuman(CompositeModule, HumanModule, _LinkContainer):
    """The Base Human Module."""

    def __init__(self, model: MainModel[Any, Any], name: str = "human"):
        HumanModule.__init__(self, model, name)
        CompositeModule.__init__(self, model, name=name)
        _LinkContainer.__init__(self)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""Base container, common methods for both the main model and the cells.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Optional,
    Type,
    Union,
)

import numpy as np
import pyproj

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from abses._bases.errors import ABSESpyError
from abses.actor import Actor, Breeds
from abses.sequences import HOW, ActorsList, Selection
from abses.tools.func import make_list

if TYPE_CHECKING:
    from abses.main import MainModel

ActorTypes: TypeAlias = Union[Type[Actor], Iterable[Type[Actor]]]
Actors: TypeAlias = Union[Actor, ActorsList, Iterable[Actor]]
UniqueID: TypeAlias = Union[str, int]


class _AgentsContainer(dict):
    """AgentsContainer for the main model."""

    def __init__(
        self,
        model: MainModel[Any, Any],
        max_len: None | int = None,
    ):
        super().__init__({b: set() for b in model.breeds})
        self._model: MainModel = model
        model._containers.append(self)
        self._max_length: Optional[int] = max_len

    def __len__(self) -> int:
        return len(self.get())

    def __str__(self) -> str:
        return "ModelAgents"

    def __repr__(self) -> str:
        strings = [f"({len(v)}){k}" for k, v in self.items()]
        return f"<{str(self)}: {'; '.join(strings)}>"

    def __contains__(self, actor: object) -> bool:
        if not isinstance(actor, Actor):
            raise TypeError(f"{type(actor)} is not a Actor.")
        return actor in self.get()

    def __call__(self, *args: Breeds, **kwargs: Breeds) -> ActorsList[Actor]:
        return self.get(*args, **kwargs)

    @property
    def crs(self) -> pyproj.CRS:
        """Returns the current CRS."""
        return self._model.nature.crs

    @property
    def model(self) -> MainModel[Any, Any]:
        """The ABSESpy model where the container belongs to."""
        return self._model

    @property
    def is_full(self) -> bool:
        """Whether the container is full."""
        return (
            False
            if self._max_length is None
            else len(self.get()) >= self._max_length
        )

    @property
    def is_empty(self) -> bool:
        """Check whether the container is empty."""
        return len(self.get()) == 0

    def check_registration(
        self, actor_cls: Type[Actor], register: bool = False
    ) -> bool:
        """Whether the breed of the actor is registered.

        Parameters:
            actor_cls:
                The class of the actor.
                If the actor is registered,
                it will be accessible from all containers
                (despite count is zero).

        Returns:
            True if the breed of the actor is registered.
            False otherwise.
        """
        flag = actor_cls.breed in self.keys()
        if not flag and register:
            self._model.agents.register(actor_cls)
        return flag

    def _check_adding_for_length(self, when_adding: int = 1) -> None:
        """Check if the container is invalid for adding the agent.

        Parameters:
            when_adding:
                The number of agents to add.

        Raises:
            ABSESpyError:
                If the container is full after adding these agents.
        """
        if self._max_length is None:
            return
        if self.has() + when_adding > self._max_length:
            e1 = f"{self} is full (maximum {self._max_length}: "
            e2 = f"Now has {self.has()}), trying to add {when_adding} more."
            raise ABSESpyError(e1 + e2)

    def get(self, breeds: Breeds = None) -> ActorsList[Actor]:
        """Get all entities of specified breeds to a list.

        Parameters:
            breeds:
                The breed(s) of entities to convert to a list.
                If None, all breeds are used.

        Returns:
            ActorsList:
                A list of entities of the specified breeds.

        Example:
            ```python
            from abses import Actor, MainModel

            class Actor1(Actor):
                pass

            class Actor2(Actor):
                pass

            model = MainModel()
            model.agents.new(Actor, singleton=True)
            model.agents.new(Actor1, num=2)
            model.agents.new(Actor2, num=3)

            model.agents.get('Actor1')
            >>> '<ActorsList: (2)Actor1>'
            model.agents.get()
            >>> '<ActorsList: (1)Actor; (2)Actor1; (3)Actor2>'
            ```
        """
        # specified breeds
        breeds = self.model.breeds if breeds is None else make_list(breeds)
        # get all available agents
        agents = {
            a
            for breed, actors in self.items()
            if breed in breeds
            for a in actors
        }
        return ActorsList(self._model, objs=agents)

    def trigger(self, *args: Any, **kwargs: Any) -> Any:
        """Trigger a function for all agents in the container.

        This method calls the `trigger` method of the list of agents in the container,
        passing the same arguments and keyword arguments received by this method.

        Parameters:
            *args:
                Positional arguments to be passed to the `trigger` method of each agent.
            **kwargs:
                Keyword arguments to be passed to the `trigger` method of each agent.

        Returns:
            In row, what the triggered function returned.
        """
        return self.get().trigger(*args, **kwargs)

    def _add_one(self, agent: Actor) -> None:
        """Add one agent to the container."""
        if not self.check_registration(agent.__class__):
            raise TypeError(
                f"Breed '{agent.breed}' not registered."
                "Is it created by `agents.new()` method?"
            )
        self[agent.breed].add(agent)

    def add(
        self,
        agents: Actors,
    ) -> None:
        """Add one or more actors to the container.

        Parameters:
            agents:
                The actor(s) to add to the container.
                It can be a single actor, a list of actors, or an iterable of actors.
            register:
                Whether to register the actor(s) if they belong to a new breed.
                If any adding breed is never registered, a TypeError will be raised.
                Once a breed is registered, it will be added to all the containers globally.
                It means, it's not necessary to register the same breed again.
                Defaults to False.

        Raises:
            TypeError:
                If a breed of the actor(s) is not registered and `register` is False.
            ABSESpyError:
                If the container is full after adding these agents.
        """
        to_add = make_list(agents)
        self._check_adding_for_length(len(to_add))
        for item in to_add:
            self._add_one(item)

    def remove(self, agent: Actor) -> None:
        """Remove the given agent from the container and the schedule.
        Generally, it stores all the agents in the model.
        Therefore, it is not recommended to use this method directly.
        Consider to use `actor.die()` instead.

        Parameters:
            agent:
                The agent (actor) to remove.

        Raises:
            ABSESpyError:
                If the agent is on a cell thus cannot be removed from the global container.
        """
        if agent.on_earth:
            raise ABSESpyError(f"{agent} is still on the earth.")
        self[agent.breed].remove(agent)
        self.model.schedule.remove(agent)

    def select(self, selection: Selection) -> ActorsList:
        """Selects the actors that match the given selection criteria.
        This method calls the `select` method of the list of actors in the container,
        passing the same selection criteria received by this method.

        Parameters:
            selection:
                The selection criteria to apply.
                Either a string or a dictionary of key-value pairs.
                Each represents agent attributes to be checked against.

        Returns:
            A list of actors that match the selection criteria.

        Example:
            ```python
            from abses import Actor, MainModel

            class Actor1(Actor):
                test = 1

            class Actor2(Actor):
                test = 'testing'

            model = MainModel()
            model.agents.new(Actor, singleton=True)
            model.agents.new(Actor1, num=2)
            model.agents.new(Actor2, num=3)

            # selecting by breed.
            model.agents.select('Actor')
            >>> '<ActorsList: (1)Actor>'

            # selecting by attribute equal expression
            model.agents.select('test == 1')
            >>> '<ActorsList: (2)Actor1>'

            # selecting by key-value pairs attribute
            model.agents.select({'test': 'testing'})
            >>> '<ActorsList: (3)Actor2>'
            ```
        """
        return self.get().select(selection)

    def has(self, breeds: Breeds = None) -> int:
        """Whether the container has the breed of agents.

        Parameters:
            breeds:
                The breed(s) of agents to search.

        Returns:
            int:
                The number of agents of the specified breed(s).

        Example:
            ```python
            from abses import Actor, MainModel

            class Actor1(Actor):
                # breed 1
                pass

            class Actor2(Actor):
                # breed 2
                pass

            model = MainModel()
            model.agents.new(Actor, singleton=True)
            model.agents.new(Actor1, num=2)
            model.agents.new(Actor2, num=3)

            model.agents.has('Actor1')
            >>> 2
            model.agents.has(['Actor1', 'Actor2'])
            >>> 5
            ```
        """
        return len(self.get(breeds=breeds))

    def apply(self, func: Callable, *args: Any, **kwargs: Any) -> np.ndarray:
        """Apply a function to all agents in the container.

        Parameters:
            func:
                The function to apply to all agents in the container.
        """
        return self.get().apply(func, *args, **kwargs)

    def item(self, how: HOW = "item", index: int = 0) -> Actor | None:
        """Retrieve one agent if possible.

        Parameters:
            how:
                The method to use to retrieve the agent.
                Can be either "only", "item", or "random".
                If "only", it will return the only agent in the container.
                In this case, the container must have only one agent.
                If more than one or no agent is found, it will raise an error.
                If "item", it will return the agent at the given index.
                If "random", it will return a randomly chosen agent.
            index:
                The index of the agent to retrieve.

        Returns:
            The agent if found, otherwise None.
        """
        return self.get().item(how=how, index=index)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import inspect
from functools import wraps
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Type,
    Union,
    cast,
)

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from abses.tools.func import camel_to_snake

if TYPE_CHECKING:
    from abses.actor import Actor

Strategy: TypeAlias = Union[str, None, bool, Number]


class Decision:
    """Decision class of actor."""

    __strategies__: Dict[str, Strategy] = {}
    _strategies: Dict[str, Strategy] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.validate_strategies(cls.__strategies__)
        cls.set_strategies(cls.__strategies__)

    def __init__(self, agent: Optional[Actor] = None) -> None:
        self._agent: Optional[Actor] = agent
        self._strategy: Any = self._setup()

    def __repr__(self) -> str:
        return f"<{self.name}: {self.now}>"

    @classmethod
    @property
    def name(cls) -> str:
        """Get the name of the decision.
        By default, this will be a snake name of class name.
        Users can custom it by assigning a class attribute `name_as`.

        Example:
            ```python
            class TestDecision(Decision):
                pass
            >>> decision = TestDecision()
            >>> decision.name
            >>> 'test_decision'

            class TestDecision(Decision):
                name_as: str = 'decision'
            >>> decision = TestDecision()
            >>> decision.name
            >>> 'decision'
            ```
        """
        default_name = camel_to_snake(cls.__name__)
        return getattr(cls, "name_as", default_name)

    @classmethod
    def validate_strategies(cls, strategies: Strategy):
        """Check if the strategies valid."""
        if not isinstance(strategies, Dict):
            raise TypeError(
                f"So far, only discrete strategies are supported, you should set '__strategies__' class attribute with a dictionary of strategies when subclassing 'Decision', got {type(strategies)} instead."
            )

    @classmethod
    def set_strategies(cls, strategies: Dict[str, Strategy]) -> None:
        """Parsing strategies and save into properties."""
        cls._strategies = strategies

    @property
    def strategies(self) -> Dict[str, Strategy]:
        """Get all strategies."""
        return self._strategies

    @property
    def agent(self) -> Optional[Actor]:
        """Decision-maker."""
        return self._agent

    @classmethod
    def making(cls, method: Callable) -> Callable:
        """A decorator makes this decision."""

        @wraps(method)
        def decorated(self: Actor, *args, **kwargs):
            cls.validate_decision_maker(self)
            result = method(self, *args, **kwargs)
            cls.validate_strategy(result)
            return result

        setattr(decorated, "__making__", cls)
        return decorated

    @classmethod
    def response(cls, strategy: Strategy) -> Callable:
        """Change the decorated function into a response methods."""

        def decorator(func) -> Callable:
            @wraps(func)
            def wrapper(self: Actor, *args, **kwargs):
                if not hasattr(self, "decisions"):
                    raise AttributeError(
                        f"{self.breed} doesn't have an attribute 'decisions'"
                    )
                if not isinstance(
                    getattr(self, "decisions"), _DecisionFactory
                ):
                    raise TypeError("Type of a decision must be decision.")
                decision_obj = self.decisions.get(cast(str, cls.name))
                if not decision_obj.has_strategy(strategy):
                    raise TypeError(
                        f"Decision '{cls.name}' doesn't have strategy {strategy}."
                    )
                return func(self, *args, **kwargs)

            setattr(wrapper, "__response__", cls)
            setattr(wrapper, "__expected__", strategy)
            return wrapper

        # decorator.__response__ = cls
        return decorator

    @classmethod
    def validate_decision_maker(cls, agent: Actor):
        """Validate maker of this decision."""
        from abses.actor import Actor

        if not isinstance(agent, Actor):
            raise TypeError(
                f"Decision maker should be an instance of 'Actor' class, got {type(agent)} instead."
            )

    @classmethod
    def validate_strategy(cls, strategy: Strategy) -> None:
        """Validate a strategy choice.

        Parameters:
            strategy:
                The strategy to validate.

        Raises:
            KeyError:
                If the strategy is not a valid choice.
        """
        if not cls.has_strategy(strategy=strategy):
            raise KeyError(
                f"Decision '{cls.__name__}' doesn't have a valid strategy {strategy}."
            )

    @classmethod
    def has_strategy(cls, strategy: Strategy) -> bool:
        """Is a specific strategy existing in this decision?

        Parameters:
            strategy:
                The strategy to validate.

        Returns:
            If the strategy exists, return True, otherwise returns False.
        """
        return strategy in cls._strategies.keys()

    @property
    def now(self) -> Any:
        """The strategy now."""
        return self._strategy

    def _setup(self) -> Any:
        if init_strategy := self.setup():  # type: ignore[error-code, func-returns-value]
            self.validate_strategy(init_strategy)
        return init_strategy

    def setup(self) -> None:
        """Overwrite to setup an initial strategy for this decision."""

    def _find_methods(self, symbol="making") -> Any:
        methods = inspect.getmembers(self.agent, predicate=inspect.ismethod)
        for _, func in methods:
            if hasattr(func, f"__{symbol}__"):
                yield func

    def _make(self) -> Any:
        for making_decision in self._find_methods("making"):
            result = making_decision()
            for response in self._find_methods("response"):
                if response.__expected__ == result:
                    response()
        return self.make()

    def make(self) -> None:
        """Overwrite this method to do something else after make decision."""


class _DecisionFactory:
    """Creating and containing decisions of an agent."""

    def __init__(
        self,
        agent: Actor,
        decisions: Optional[Iterable[Type[Decision]]] = None,
    ) -> None:
        self.agent: Actor = agent
        self._decisions: Dict[str, Decision] = {}
        if decisions is None:
            decisions = []
        self.parse_decisions(decisions)

    @property
    def agent(self) -> Actor:
        """Decision-maker, who has these decisions."""
        return self._agent

    @agent.setter
    def agent(self, agent: Actor) -> None:
        from abses.actor import Actor

        if not isinstance(agent, Actor):
            raise TypeError(
                f"Agent must be an instance of 'Actor' class, got {type(agent)} instead."
            )
        self._agent = agent

    def parse_decisions(self, decisions: Iterable[Type[Decision]]):
        """Parse decisions and save into the container.

        Parameters:
            decisions:
                Iterable `Decision` class.

        Raises:
            TypeError:
                If the input decision is not a subclass of `Decision`.
        """
        for d in decisions:
            if not issubclass(d, Decision):
                raise TypeError(
                    f"Decision must be an subclass of 'Decision' class, got {type(d)} instead."
                )
            obj = d(agent=self.agent)
            self._decisions[obj.name] = obj

    def __getattribute__(self, __name: str) -> Any:
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        if __name in self._decisions:
            return self._decisions[__name]
        return super().__getattribute__(__name)

    def keys(self) -> List[str]:
        """Get all decision names."""
        return list(self._decisions.keys())

    def get(self, name: str) -> Decision:
        """Get a decision by its name."""
        if name not in self._decisions:
            raise KeyError(
                f"Decision '{name}' doesn't exist in {self.keys()}."
            )
        return self._decisions[name]

    def making(self) -> None:
        """Making decisions."""
        for d in self._decisions.values():
            d._make()



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""Logging module.
"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Optional

from loguru import logger

if TYPE_CHECKING:
    from abses import Experiment, MainModel

logger.remove(0)
FORMAT = "[{time:HH:mm:ss}][{level}][{module}] {message}\n"


def formatter(record) -> str:
    """Customize formatter."""
    return "{message}\n" if record["extra"].get("no_format") else FORMAT


logger.add(
    sys.stderr,
    format=formatter,
    level="WARNING",
    colorize=True,
)


def log_session(title: str, msg: str = ""):
    """logging a new module session."""
    first_line = "\n" + "=" * 20 + "\n"
    center_line = f"  {title}  ".center(20, "-")
    end_line = "\n" + "=" * 20 + "\n"
    ending = "".center(20, "-")
    logger.bind(no_format=True).info(
        first_line + center_line + end_line + msg + ending
    )


def setup_logger_info(
    exp: Optional[Experiment] = None,
    # model: Optional[MainModel] = None,
):
    """Set up logger."""
    line_equal = "".center(40, "=") + "\n"
    line_star = "".center(40, "·") + "\n"
    content = "  ABSESpy Framework  ".center(40, "*") + "\n"
    msg = line_equal + line_star + content + line_star + line_equal
    logger.bind(no_format=True).info(msg)
    is_exp_env = exp is not None
    logger.bind(no_format=True).info(f"Exp environment: {is_exp_env}\n")



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
每一个世界里的斑块
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Union

from mesa_geo.raster_layers import RasterBase
from pyproj import CRS

from abses._bases.errors import ABSESpyError
from abses._bases.objects import _BaseObj
from abses.container import _CellAgentsContainer
from abses.links import TargetName, _LinkNodeCell

if TYPE_CHECKING:
    from abses.main import MainModel
    from abses.nature import PatchModule
    from abses.sequences import ActorsList

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

Pos: TypeAlias = Tuple[int, int]


def raster_attribute(
    func: Callable[[Any], Union[str | int | float]]
) -> property:
    """Turn the method into a property that the patch can extract.
    Example:
        ```
        class TestCell(Cell):
            @raster_attribute
            def test:
                return 1

        # Using this test cell to create a PatchModule.
        module = PatchModule.from_resolution(
            model=MainModel(),
            shape=(3, 3),
            cell_cls=TestCell,
        )

        # now, the attribute 'test' of TestCell can be accessible in the module, as spatial data (i.e., raster layer).

        >>> module.cell_properties
        >>> set('test')

        >>> array = module.get_raster('test')
        >>> np.array([
            [1, 1, 1],
            [1, 1, 1],
            [1, 1, 1],
        ])
        ```
    """
    setattr(func, "is_decorated", True)
    return property(func)


class PatchCell(_LinkNodeCell, _BaseObj):
    """A patch cell of a `RasterLayer`.
    Subclassing this class to create a custom cell.
    When class attribute `max_agents` is assigned,
    the `agents` property will be limited to the number of agents.

    Attributes:
        agents:
            The agents located at here.
        layer:
            The `RasterLayer` where this `PatchCell` belongs.
    """

    max_agents: Optional[int] = None

    def __init__(self, layer, indices: Pos):
        _BaseObj.__init__(self, model=layer.model, observer=True)
        _LinkNodeCell.__init__(self)
        self.indices = indices
        self._set_layer(layer=layer)

    def __repr__(self) -> str:
        return f"<Cell at {self.layer}[{self.indices}]>"

    @classmethod
    def __attribute_properties__(cls) -> set[str]:
        """Properties that should be found in the `RasterLayer`.

        Users should decorate a property attribute when subclassing `PatchCell` to make it accessible in the `RasterLayer`.
        """
        return {
            name
            for name, method in cls.__dict__.items()
            if isinstance(method, property)
            and getattr(method.fget, "is_decorated", False)
        }

    @property
    def layer(self) -> PatchModule:
        """`RasterLayer` where this `PatchCell` belongs."""
        if self._layer is None:
            raise ABSESpyError(
                "PatchCell must belong to a layer."
                f"However, {self} has no layer."
                "Did you create this cell in the correct way?"
            )
        return self._layer

    @property
    def agents(self) -> _CellAgentsContainer:
        """The agents located at here."""
        return self._agents

    @property
    def coordinate(self) -> Tuple[float, float]:
        """The position of this cell."""
        row, col = self.indices
        return self.layer.transform_coord(row=row, col=col)

    @property
    def geo_type(self) -> str:
        """Return the geo_type"""
        return "Cell"

    @property
    def crs(self) -> Optional[CRS]:
        """Return the crs of this cell."""
        return self.layer.crs

    def _set_layer(self, layer: PatchModule) -> None:
        if not isinstance(layer, RasterBase):
            raise TypeError(f"{type(layer)} is not valid layer.")
        # set layer property
        self._layer = layer
        # set layer's model as the model
        self.model: MainModel[Any, Any] = layer.model
        # set agents container
        self._agents = _CellAgentsContainer(
            layer.model, cell=self, max_len=getattr(self, "max_agents", None)
        )

    def get(self, attr: str, target: Optional[TargetName] = None) -> Any:
        """Gets the value of an attribute or registered property.
        Automatically update the value if it is the dynamic variable of the layer.

        Parameters:
            attr:
                The name of attribute to get.

        Returns:
            Any:
                The value of the attribute.

        Raises:
            AttributeError:
                Attribute value of the associated patch cell.
        """
        if attr in self.layer.dynamic_variables:
            self.layer.dynamic_var(attr_name=attr)
        return super().get(attr=attr, target=target)

    def neighboring(
        self,
        moore: bool = False,
        radius: int = 1,
        include_center: bool = False,
        annular: bool = False,
    ) -> ActorsList[_LinkNodeCell]:
        """Get the grid around the patch."""
        return self.layer.get_neighborhood(
            self.indices,
            moore=moore,
            radius=radius,
            include_center=include_center,
            annular=annular,
        )



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
基础类实现了观察者模式。

让所有模型的更改能够实时分配给所有观察者。

The base class implements the observer pattern.
It allows to observe all attribute changes in the main model.
"""

from __future__ import annotations

from abc import ABCMeta
from typing import List, Optional, Set, Union

from abses.tools.func import make_list


class _Notice:
    """Notice class for the observer pattern."""

    __glob_vars__: Set[str] = set()

    def __init__(self, observer: Optional[_Observer] = None):
        self.observers: Set[_Observer] = set()
        self._glob_vars = set(self.__glob_vars__)
        if observer is not None:
            self.attach(observer)

    def __repr__(self) -> str:
        return (
            f"<Noticing {self.glob_vars} to {len(self.observers)} observers>"
        )

    @property
    def glob_vars(self) -> List[str]:
        """
        The global variable list, where the variables are automatically updated to all observers.
        """
        return sorted(self._glob_vars)

    def add_glob_vars(self, value: Union[str, List[str]]) -> None:
        """Add new global variables.

        Parameters:
            value:
                str or list of str
                The new global variable(s) to be added.
        """
        new_vars = set(make_list(value))
        for var in new_vars:
            if not hasattr(self, var):
                raise AttributeError(
                    f"{var} is not a variable in {self.__class__}."
                )
            self._glob_vars.add(var)
        self.notify()

    def attach(self, observer: _Observer) -> None:
        """Add a new observer."""
        self.observers.add(observer)
        observer.notification(self)

    def detach(self, observer: _Observer) -> None:
        """Detach an observer."""
        self.observers.remove(observer)

    def notify(self) -> None:
        """Notify all observers."""
        for observer in self.observers:
            observer.notification(self)


class _Observer(metaclass=ABCMeta):
    """
    Observer is responsible for receiving notifications from the main model.
    """

    def notification(self, notice: _Notice):
        """When the main model changes, the observer will receive a notification."""
        for var in notice.glob_vars:
            value = getattr(notice, var)
            setattr(self, var, value)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""数据收集器
"""
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    cast,
)

import numpy as np
import pandas as pd
from loguru import logger

if TYPE_CHECKING:
    from abses.actor import Actor
    from abses.main import MainModel
    from abses.sequences import ActorsList
    from abses.time import TimeDriver

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

Reporter: TypeAlias = Callable[..., Any]
ReporterDict: TypeAlias = Dict[str, Reporter]
ReportType: TypeAlias = Literal["model", "agents", "final"] | str


def _getattr_to_reporter(
    attribute_name: str,
) -> Callable[..., Any]:
    """获取属性的报告函数"""

    def attr_reporter(obj: Actor | MainModel):
        return getattr(obj, attribute_name, None)

    return attr_reporter


def _func_reporter(reporter: List) -> Callable[..., Any]:
    """函数报告器"""
    func, params = reporter[0], reporter[1]

    def func_reporter(agent: Actor):
        return func(agent, *params)

    return func_reporter


def clean_to_reporter(
    reporter: Reporter,
    *args,
    **kwargs,
) -> Callable[..., Any]:
    """将字符串转换为函数"""
    if isinstance(reporter, str):
        reporter = _getattr_to_reporter(attribute_name=reporter)
    elif isinstance(reporter, Iterable):
        reporter = _func_reporter(reporter)
    elif callable(reporter):
        reporter = _func_reporter([reporter, args, kwargs])
    return reporter


class ABSESpyDataCollector:
    """ABSESpyDataCollector, adapted from DataCollector of `mesa`."""

    def __init__(self, reports: Dict[ReportType, Dict[str, Reporter]]):
        self.model_reporters: Dict[str, Reporter] = {}
        self.final_reporters: Dict[str, Reporter] = {}
        self.agent_reporters: Dict[str, Dict[str, Reporter]] = {}

        self._agent_records: Dict[str, List[pd.DataFrame]] = {}
        self.model_vars: Dict[str, List[Any]] = {}

        self.add_reporters("model", reports.get("model", {}))
        self.add_reporters("agents", reports.get("agents", {}))
        self.add_reporters("final", reports.get("final", {}))

    def add_reporters(
        self,
        item: ReportType,
        reporters: ReporterDict,
    ) -> None:
        """Add a dictionary of new reporters."""
        # 处理列表？
        # if isinstance(reporters, (tuple, list)):
        #     reporters = {name: name for name in reporters}
        if item == "model":
            for name, reporter in reporters.items():
                self._new_model_reporter(name=name, reporter=reporter)
            return
        if item == "final":
            for name, reporter in reporters.items():
                self.final_reporters[name] = clean_to_reporter(reporter)
            return
        if item == "agents":
            for breed, tmp_reporters in reporters.items():
                self.add_reporters(
                    item=breed, reporters=cast(ReporterDict, tmp_reporters)
                )
            return
        for name, reporter in reporters.items():
            self._new_agent_reporter(breed=item, name=name, reporter=reporter)

    def _new_model_reporter(self, name: str, reporter: Reporter) -> None:
        """Add a new model-level reporter to collect data.

        Parameters:
            name:
                Name of the model level variable to collect.
            reporter:
                Attribute string,
                or function object that returns the variable.
        """
        self.model_reporters[name] = clean_to_reporter(reporter)
        self.model_vars[name] = []

    def _record_a_breed_of_agents(
        self, time: TimeDriver, breed: str, agents: ActorsList[Actor]
    ) -> None:
        """记录某一组的数据"""
        result = {
            "AgentID": agents.array("unique_id"),
            "Step": np.repeat(time.tick, len(agents)),
            "Time": np.repeat(time.dt, len(agents)),
        }
        for name, reporter in self.agent_reporters[breed].items():
            result[name] = agents.apply(reporter)
        self._agent_records[breed].append(result)

    def _record_agents(self, model: MainModel) -> None:
        """记录所有的Agents"""
        for breed in model.breeds:
            if breed not in self.agent_reporters:
                continue
            if breed not in self._agent_records:
                self._agent_records[breed] = []
            agents = model.agents.get(breed)
            self._record_a_breed_of_agents(model.time, breed, agents)

    def _new_agent_reporter(
        self, breed: str, name: str, reporter: Reporter
    ) -> None:
        """添加新的 Agent Reporter"""
        if breed not in self.agent_reporters:
            self.agent_reporters[breed] = {}
        self.agent_reporters[breed][name] = clean_to_reporter(
            reporter=reporter
        )

    def get_model_vars_dataframe(self):
        """Create a pandas DataFrame from the model variables.

        The DataFrame has one column for each model variable, and the index is
        (implicitly) the model tick.
        """
        # Check if self.model_reporters dictionary is empty, if so raise warning
        if not self.model_reporters:
            logger.warning(
                "No model reporters have been defined"
                "returning empty DataFrame."
            )

        return pd.DataFrame(self.model_vars)

    def get_agent_vars_dataframe(
        self, breed: Optional[str] = None
    ) -> pd.DataFrame:
        """获取某种 Agents 的 DataFrame"""
        if breed is None:
            return {
                breed: self.get_agent_vars_dataframe(breed)
                for breed in self.agent_reporters
            }
        if not self.agent_reporters:
            logger.warning(
                "No agent reporters have been defined in the DataCollector."
            )
        if results := self._agent_records.get(breed):
            return pd.concat([pd.DataFrame(res) for res in results])
        return pd.DataFrame()

    def get_final_vars_report(self, model: MainModel) -> pd.DataFrame:
        """Report at the end of this model."""
        if not self.final_reporters:
            logger.warning(
                "No final reporters have been defined"
                "returning empty DataFrame."
            )
        return {var: func(model) for var, func in self.final_reporters.items()}

    def collect(self, model: MainModel):
        """Collect all the data for the given model object."""

        if self.model_reporters:
            for var, func in self.model_reporters.items():
                self.model_vars[var].append(func(model))

        if self.agent_reporters:
            self._record_agents(model)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
This script is used to manipulate actors' movements.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Optional, Tuple, cast

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from mesa.space import Coordinate
from mesa_geo.raster_layers import RasterBase

from abses._bases.errors import ABSESpyError
from abses.links import _LinkNodeCell

if TYPE_CHECKING:
    from abses.actor import Actor
    from abses.cells import PatchCell
    from abses.nature import PatchModule

MovingDirection: TypeAlias = Literal[
    "left",
    "right",
    "up",
    "down",
    "up left",
    "up right",
    "down left",
    "down right",
]


def _get_layer_and_position(
    pos: Coordinate | PatchCell, layer: Optional[PatchModule] = None
) -> Tuple[Optional[PatchModule], Coordinate]:
    if isinstance(cast("PatchCell", pos), _LinkNodeCell):
        if layer is not None and layer is not pos.layer:
            raise ABSESpyError(
                "The input layer is not consistent with the cell's layer."
            )
        return pos.layer, pos.indices
    if isinstance(pos, tuple) and len(pos) == 2:
        return layer, pos
    raise TypeError(f"Invalid position type {pos}.")


def _put_agent_on_cell(agent: Actor, cell: PatchCell) -> None:
    """
    This function is used to put an agent on a cell.
    """
    if not isinstance(cell, _LinkNodeCell):
        raise TypeError(
            f"Agent must be put on a `abses.PatchCell`, instead of {type(cell)}"
        )
    # leave the old cell.
    if agent.on_earth:
        agent.move.off()
    # before moving to the new cell, agent may do something
    keep_moving = agent.moving(cell=cell)
    if keep_moving is False:
        return
    # put the agent on the new cell after check.
    cell.agents.add(agent)
    agent.at = cell


def move_agent_to(
    agent: Actor,
    layer: PatchModule,
    pos: Coordinate | _LinkNodeCell,
) -> None:
    """Move an Actor to another position of this layer.

    Parameters:
        agent:
            The actor to operate.
        position:
            The position to put on.

    Raises:
        ABSESpyError:
            If the agent is not located at this layer before moving.

    Returns:
        If the operation is successful.
    """
    expected_layer, pos = _get_layer_and_position(pos)
    if not isinstance(layer, RasterBase):
        raise TypeError(f"{layer} is not valid 'PatchModule'.")
    if expected_layer and expected_layer is not layer:
        raise ABSESpyError(
            f"{pos} expects operation on the layer {expected_layer}, got input {layer}."
        )
    if layer.out_of_bounds(pos):
        raise ValueError(f"Position {pos} is out of bounds.")
    cell = layer.array_cells[pos[0], pos[1]]
    _put_agent_on_cell(agent, cell)


class _Movements:
    """
    This class is used to manipulate actors' movements.
    """

    def __init__(self, actor: Actor) -> None:
        self.actor = actor
        self.model = actor.model
        self.seed = actor.unique_id
        # self.direction = actor.direction

    @property
    def layer(self) -> Optional[PatchModule]:
        """The current layer of the operating actor."""
        return self.actor.layer

    def _operating_layer(
        self, layer: Optional[PatchModule]
    ) -> Optional[PatchModule]:
        """
        This method is used to check if the input layer is consistent with the actor's layer.
        """
        if self.layer is None and layer is None:
            raise ABSESpyError("No operating layer is specified.")
        if self.layer is layer or self.layer is None:
            return layer
        if layer is None:
            return self.layer
        raise ABSESpyError(
            "The input layer is not consistent with the actor's layer."
        )

    def to(
        self,
        pos: PatchCell | Coordinate | Literal["random"],
        layer: Optional[PatchModule] = None,
    ) -> None:
        """
        Move the actor to a specific location.

        Parameters:
            pos:
                The position to move to.
                If position is a Coordinate -a tuple of (row, col),
                it will be moved to the same layer.
                If pos
            layer:
                The layer where the actor is located.

        Raises:
            ABSESpyError:
                If the input layer is not consistent with the actor's layer.
                If the position is out of bounds.
                Or, if the pos is coordinate without layer.
        """
        if isinstance(pos, str) and pos == "random":
            # 随机分配一个该图层的位置
            operating_layer = self._operating_layer(layer=layer)
            assert operating_layer is not None
            pos = operating_layer.select().random.choice()
        else:
            # 检查这个位置的类型，返回图层和位置
            layer, pos = _get_layer_and_position(pos, layer=layer)
            operating_layer = self._operating_layer(layer=layer)
            assert operating_layer is not None
        move_agent_to(self.actor, layer=operating_layer, pos=pos)

    def off(self) -> None:
        """Remove the actor from the world."""
        if self.actor.at is None:
            raise ABSESpyError("The actor is not located on a cell.")
        if self.actor.on_earth:
            container = self.actor.at.agents
            if container is None:
                raise ABSESpyError("The actor is not located on a cell.")
            container.remove(self.actor)
        del self.actor.at

    def by(self, direction: MovingDirection, distance: int = 1) -> None:
        """Move the actor by a specific distance.

        Parameters:
            direction:
                The direction to move.
                It should be a direction string such as:
                "left", "right", "up", "down", "up left", "up right", "down left", "down right".
            distance:
                The distance to move toward the direction.

        Raises:
            ABSESpyError:
                If the actor is not located on a cell, thus cannot move.
            ValueError:
                If the direction is invalid.
        """
        if (self.actor.at is None) or (self.layer is None):
            raise ABSESpyError(
                "The actor is not located on a cell, thus cannot move."
            )
        old_row, old_col = self.actor.at.indices
        if direction == "left":
            new_indices = (old_row, old_col - distance)
        elif direction == "right":
            new_indices = (old_row, old_col + distance)
        elif direction == "up":
            new_indices = (old_row - distance, old_col)
        elif direction == "down":
            new_indices = (old_row + distance, old_col)
        elif direction in {"up left", "left up"}:
            new_indices = (old_row - distance, old_col - distance)
        elif direction in {"up right", "right up"}:
            new_indices = (old_row - distance, old_col + distance)
        elif direction in {"down left", "left down"}:
            new_indices = (old_row + distance, old_col - distance)
        elif direction in {"down right", "right down"}:
            new_indices = (old_row + distance, old_col + distance)
        else:
            raise ValueError(f"Invalid direction {direction}.")
        cell = self.layer.array_cells[new_indices[0]][new_indices[1]]
        self.actor.move.to(cell)

    def random(self, prob: Optional[str] = None, **kwargs: Any) -> None:
        """Move the actor to a random location nearby.

        Parameters:
            prob:
                The probability to select a cell.
            kwargs:
                Passing keyword args to `PatchCell.neighboring`,
                used to select neighboring cells.
        """
        if self.actor.at is None:
            raise ABSESpyError("The actor is not located on a cell.")
        cells = self.actor.at.neighboring(**kwargs)
        self.actor.move.to(cells.random.choice(prob=prob))



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""Viz natural module.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, cast

from matplotlib.axes import Axes

from abses.tools.func import with_axes
from abses.tools.viz import COLOR_BAR

if TYPE_CHECKING:
    from abses.actor import Actor
    from abses.nature import PatchModule


class _VizNature:
    """Visualize the nature module"""

    def __init__(self, module: PatchModule):
        self.module = module
        self.model = module.model

    @with_axes
    def show(
        self,
        attr: Optional[str] = None,
        ax: Optional[Axes] = None,
        with_actors: bool = True,
        scatter_kwargs: Optional[Dict[str, Any]] = None,
        legend_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Axes:
        """Show the nature module"""
        if scatter_kwargs is None:
            scatter_kwargs = {}
        if legend_kwargs is None:
            legend_kwargs = {}
        # because of `with_axes`, it must be a valid Axes object
        ax = cast(Axes, ax)
        if attr is None:
            xda = self.module.xda
            xda.plot(ax=ax, add_colorbar=False, alpha=0.8)
        else:
            xda = self.module.get_xarray(attr)
            xda.plot(ax=ax, cbar_kwargs=COLOR_BAR, alpha=0.8)
        if self.model.breeds and with_actors:
            self.model.actors.plot.show(ax=ax, **scatter_kwargs)
        ax.axes.set_aspect("equal")
        if self.model.breeds:
            ax.legend(**legend_kwargs)
        return ax



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
This module contains some commonly used regular expressions for checking names.
这个模块里存储一些检查名称常用的正则表达式。
"""
import re

# 模块名称应该符合蛇形命名法，且不能以下划线开头
# Module name is snake case and should not start with an underscore
# https://www.cnblogs.com/mrlonely2018/p/15650605.html
MODULE_NAME = re.compile(r"[a-z][a-z0-9_]*")

# 类名应该符合驼峰命名法
# Class name should be in camel case
CAMEL_NAME = re.compile(r"(?<!^)(?=[A-Z])")



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""ActorsList is a sequence of actors.
It's used to manipulate the actors quickly in batch.
"""

from __future__ import annotations

from collections.abc import Iterable
from functools import cached_property, partial
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Sized,
    TypeVar,
    Union,
    cast,
    overload,
)

import pandas as pd
from loguru import logger
from pyproj import CRS

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

import geopandas as gpd
import mesa_geo as mg
import numpy as np
from numpy.typing import NDArray

from abses._bases.errors import ABSESpyError
from abses.random import ListRandom
from abses.selection import selecting
from abses.tools.func import make_list
from abses.viz.viz_actors import _VizNodeList

if TYPE_CHECKING:
    from abses._bases.base_container import UniqueID
    from abses.actor import Actor, GeoType, TargetName
    from abses.links import _LinkNode
    from abses.main import MainModel

Selection: TypeAlias = Union[str, Iterable[bool], Dict[str, Any]]
HOW: TypeAlias = Literal["only", "random", "item"]
Link = TypeVar("Link", bound="_LinkNode")


def get_only_agent(agents: ActorsList) -> Actor:
    """Select one agent"""
    if len(agents) == 0:
        raise ValueError("No agent found.")
    if len(agents) == 1:
        return agents[0]
    raise ValueError("More than one agent.")


class ActorsList(List[Link], Generic[Link]):
    """A list of actors in an agent-based model."""

    def __init__(
        self, model: MainModel[Any, Any], objs: Iterable[Link] = ()
    ) -> None:
        super().__init__(objs)
        self._model = model

    def __repr__(self):
        results = [f"({len(v)}){k}" for k, v in self.to_dict().items()]
        return f"<ActorsList: {'; '.join(results)}>"

    def __eq__(self, other: Iterable[Any]) -> bool:
        return (
            all(actor in other for actor in self)
            if self._is_same_length(cast(Sized, other))
            else False
        )

    @overload
    def __getitem__(self, other: int) -> Link:
        ...

    @overload
    def __getitem__(self, index: slice) -> ActorsList[Link]:
        ...

    def __getitem__(self, index):
        results = super().__getitem__(index)
        return (
            ActorsList(self._model, results)
            if isinstance(index, slice)
            else results
        )

    def _is_same_length(self, length: Sized, rep_error: bool = False) -> bool:
        """Check if the length of input is as same as the number of actors."""
        if not hasattr(self, "__len__"):
            raise ValueError(f"{type(length)} object is not iterable.")
        if len(length) != len(self):
            if rep_error:
                raise ValueError(
                    f"Length of the input {len(length)} mismatch {len(self)} actors."
                )
            return False
        return True

    @cached_property
    def random(self) -> ListRandom:
        """Random module"""
        return ListRandom(actors=self, model=self._model)

    @cached_property
    def plot(self) -> _VizNodeList:
        """Plotting module"""
        return _VizNodeList(self._model, self)

    def to_dict(self) -> Dict[str, ActorsList[Link]]:
        """Convert all actors in this list to a dictionary like {breed: ActorList}.

        Returns:
            key is the breed of actors, and values are corresponding actors.
        """
        dic: Dict[str, ActorsList[Link]] = {}
        for actor in iter(self):
            breed = actor.breed
            if breed not in dic:
                dic[breed] = ActorsList(self._model, [actor])
            else:
                dic[breed].append(actor)
        return dic

    def _subset(
        self, geo_type: Optional[GeoType | bool] = None
    ) -> ActorsList[Link]:
        """Returns dataset for plotting."""
        if geo_type is None:
            return self
        if isinstance(geo_type, bool):
            selection: Dict[str, Any] = {"on_earth": geo_type}
        elif geo_type in ("Point", "Shape"):
            selection = {"geo_type": geo_type}
        return self.select(selection)

    def select(
        self,
        selection: Optional[Selection] = None,
        geo_type: Optional[GeoType] = None,
    ) -> ActorsList[Link]:
        """
        Returns a new :class:`ActorList` based on `selection`.

        Parameters:
            selection:
                List with same length as the agent list.
                Positions that return True will be selected.
            geo_type:
                Type of Actors' Geometry.

        Returns:
            A subset containing.
        """
        actors = self._subset(geo_type=geo_type)
        if selection is None:
            return actors
        if isinstance(selection, (str, dict)):
            bool_ = [selecting(actor, selection) for actor in actors]
        elif isinstance(selection, (list, tuple, np.ndarray)):
            bool_ = make_list(selection)
        else:
            raise TypeError(f"Invalid selection type {type(selection)}")
        selected = [a for a, s in zip(actors, bool_) if s]
        return ActorsList(self._model, selected)

    def ids(self, ids: Iterable[UniqueID] | UniqueID) -> ActorsList[Link]:
        """Subsets ActorsList by a `ids`.

        Parameters:
            ids:
                an iterable id list. List[id], ID is an attr of agent obj.

        Returns:
            ActorList: A subset of origin agents list.
        """
        ids = make_list(ids)
        return self.select([agent.unique_id in ids for agent in self])

    def better(
        self, metric: str, than: Optional[Union[Number, Actor]] = None
    ) -> ActorsList[Link]:
        """
        Selects the elements of the sequence that are better than a given value or actor
        based on a specified metric.

        Parameters:
            metric:
                The name of the attribute to use as the metric for comparison.
            than:
                The value or actor to compare against. If None, selects the elements with the
                highest value for the specified metric. If a number, selects the elements with
                a value greater than the specified number. If an Actor, selects the elements
                with a value greater than the specified Actor's value for the specified metric.

        Returns:
            A new sequence containing the selected elements.

        Raises:
            ABSESpyError:
                If the `than` parameter is not a Number or an Actor.

        Notes:
            This method compares the values of the specified metric for all elements in the
            sequence and selects the elements that are better than the specified value or actor.
            The comparison is based on the greater than operator (>) for numbers and the
            difference between the values for actors.
        """
        metrics = self.array(attr=metric)
        if than is None:
            return self.select(metrics == max(metrics))
        if isinstance(than, Number):
            return self.select(metrics > than)
        if isinstance(than, mg.GeoAgent):
            diff = self.array(metric) - getattr(than, metric)
            return self.select(diff > 0)
        raise ABSESpyError(f"Invalid than type {type(than)}.")

    def update(self, attr: str, values: Iterable[Any]) -> None:
        """Update the specified attribute of each agent in the sequence with the corresponding value in the given iterable.

        Parameters:
            attr:
                The name of the attribute to update.
            values:
                An iterable of values to update the attribute with. Must be the same length as the sequence.

        Raises:
            ValueError:
                If the length of the values iterable does not match the length of the sequence.
        """
        self._is_same_length(cast(Sized, values), rep_error=True)
        for agent, val in zip(self, values):
            setattr(agent, attr, val)

    def split(self, where: NDArray[Any]) -> List[ActorsList[Link]]:
        """Split agents into N+1 groups.

        Parameters:
            where:
                indexes [size=N] denotes where to split.

        Returns:
            np.ndarray: N+1 groups: agents array
        """
        split = np.hsplit(np.array(self), where)
        return [ActorsList(self._model, group) for group in split]

    def array(self, attr: str) -> np.ndarray:
        """Convert the specified attribute of all actors to a numpy array.

        Parameters:
            attr:
                The name of the attribute to convert to a numpy array.

        Returns:
            A numpy array containing the specified attribute of all actors.
        """
        return np.array([getattr(actor, attr) for actor in self])

    def trigger(self, func_name: str, *args: Any, **kwargs: Any) -> np.ndarray:
        """Call a method with the given name on all actors in the sequence.

        Parameters:
            func_name:
                The name of the method to call on each actor.
            *args:
                Positional arguments to pass to the method.
            **kwargs:
                Keyword arguments to pass to the method.

        Returns:
            An array of the results of calling the method on each actor.
        """
        results = [
            getattr(actor, func_name)(*args, **kwargs) for actor in iter(self)
        ]
        return np.array(results)

    def apply(self, ufunc: Callable, *args: Any, **kwargs: Any) -> np.ndarray:
        """Apply ufunc to all actors in the sequence.

        Parameters:
            ufunc:
                The function to apply to each actor.
            *args:
                Positional arguments to pass to the function.
            **kwargs:
                Keyword arguments to pass to the function.

        Returns:
            An array of the results of applying the function to each actor.
        """
        func = partial(ufunc, *args, **kwargs)
        return np.array(list(map(func, self)))

    def get(
        self,
        attr: str,
        target: Optional[TargetName] = None,
        how: HOW = "only",
        default: Optional[Any] = None,
    ) -> Any:
        """Retrieve the attribute of an either specified or randomly chosen agent.

        Parameters:
            attr:
                The name of the attribute to retrieve.
            how:
                The method to use to retrieve the attribute. Can be either "only" or "random".

        Returns:
            The attribute of the specified agent.
        """
        if agent := self.item(how=how, index=0):
            return agent.get(attr, target=target)
        if default is not None:
            return default
        raise ValueError("No agent found or default value.")

    def set(
        self, attr: str, value: Any, target: Optional[TargetName] = None
    ) -> None:
        """Set the attribute of all agents in the sequence to the specified value.

        Parameters:
            attr:
                The name of the attribute to set.
            value:
                The value to set the attribute to.
        """
        for agent in iter(self):
            agent.set(attr, value, target=target)

    def item(self, how: HOW = "item", index: int = 0) -> Optional[Link]:
        """Retrieve one agent if possible.

        Parameters:
            how:
                The method to use to retrieve the agent.
                Can be either "only", "item", or "random".
                If "only", it will return the only agent in the container.
                In this case, the container must have only one agent.
                If more than one or no agent is found, it will raise an error.
                If "item", it will return the agent at the given index.
                If "random", it will return a randomly chosen agent.
            index:
                The index of the agent to retrieve.

        Returns:
            The agent if found, otherwise None.
        """
        if how == "only":
            return get_only_agent(self)
        if how == "random":
            actor = self.random.choice(when_empty="return None")
            return cast(Optional["Actor"], actor)
        if how == "item":
            return self[index] if len(self) > index else None
        raise ValueError(f"Invalid how method '{how}'.")

    def _check_crs_consistent(self) -> CRS:
        crs_set = set(self.array("crs"))
        if None in crs_set:
            logger.warning("Some agents don't have a crs.")
            crs_set.remove(None)
        if len(crs_set) > 1:
            raise ValueError(f"More than one crs: {crs_set}.")
        if not crs_set:
            logger.warning("No crs when init a GeoDataFrame.")
        return crs_set.pop()

    @overload
    def summary(self, geometry: bool = True, **kwargs) -> gpd.GeoDataFrame:
        ...

    @overload
    def summary(self, geometry: bool = False, **kwargs) -> pd.DataFrame:
        ...

    def summary(
        self, geometry: bool = False, **kwargs
    ) -> pd.DataFrame | gpd.GeoDataFrame:
        """Returns a summarized dataframe of the actors."""
        if len(self) == 0:
            raise ValueError("No actors to retrieve summary information.")
        df = pd.concat([actor.summary(**kwargs) for actor in self], axis=1).T
        if geometry:
            crs = self._check_crs_consistent()
            df["geometry"] = self.array("geometry")
            return gpd.GeoDataFrame(df, geometry="geometry", crs=crs)
        return df



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""Handling some example datasets.
"""
from importlib import resources
from pathlib import Path
from typing import Literal

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias


DATA_LIST = [
    "farmland.tif",
    "precipitation.nc",
    "irr_lands.csv",
    "YR_cities.zip",
]
DataChoices: TypeAlias = Literal[
    "farmland.tif",
    "precipitation.nc",
    "irr_lands.csv",
    "YR_cities.zip",
]


def load_data(filename: DataChoices) -> Path:
    """Load data from the package."""
    if filename not in DATA_LIST:
        raise FileNotFoundError(
            f"Cannot load data {filename},"
            f"Example Datasets include: {DATA_LIST}."
        )
    return Path(str(resources.files("data") / filename))



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
行动者容器，集中保存行动者。
Container for actors.
"""

from __future__ import annotations

import contextlib
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

with contextlib.suppress(ImportError):
    import networkx as nx
try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

import geopandas as gpd
from loguru import logger
from shapely.geometry.base import BaseGeometry

from abses._bases.base_container import _AgentsContainer
from abses._bases.errors import ABSESpyError
from abses.actor import Actor
from abses.links import get_node_unique_id
from abses.sequences import ActorsList
from abses.tools.func import IncludeFlag, clean_attrs, make_list

if TYPE_CHECKING:
    from abses.cells import PatchCell
    from abses.main import MainModel

ActorTypes: TypeAlias = Union[Type[Actor], Iterable[Type[Actor]]]
Actors: TypeAlias = Union[Actor, ActorsList, Iterable[Actor]]
UniqueID: TypeAlias = Union[str, int]
UniqueIDs: TypeAlias = List[Optional[UniqueID]]


class _ModelAgentsContainer(_AgentsContainer):
    """AgentsContainer for the MainModel."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._unique_ids: Dict[str, int] = {}

    def _check_unique_id(
        self, unique_id: Optional[UniqueID], breed: str
    ) -> UniqueID:
        if unique_id is None:
            unique_id = f"{breed}[{self._unique_ids[breed]}]"
            self._unique_ids[breed] += 1
        if not isinstance(unique_id, (int, str)):
            raise ABSESpyError(
                f"Unique ID should be an integer or string, got {type(unique_id)}."
            )
        if unique_id in set(self.get().array("unique_id")):
            raise ABSESpyError(f"Unique ID {unique_id} already exists.")
        return unique_id

    def _check_crs(self, gdf: gpd.GeoDataFrame) -> bool:
        if gdf.crs:
            gdf.to_crs(self.crs, inplace=True)
        else:
            gdf.set_crs(self.crs, inplace=True)
        return self.crs == gdf.crs

    def _new_one(
        self,
        unique_id: Optional[UniqueID] = None,
        geometry: Optional[BaseGeometry] = None,
        agent_cls: type[Actor] = Actor,
        **kwargs,
    ) -> Actor:
        if geometry and not isinstance(geometry, BaseGeometry):
            raise TypeError("Geometry must be a Shapely Geometry")
        unique_id = self._check_unique_id(unique_id, breed=agent_cls.breed)
        agent = agent_cls(
            unique_id=unique_id,
            model=self.model,
            geometry=geometry,
            crs=self.model.nature.crs,
            **kwargs,
        )
        self.add(agent)
        self.model.schedule.add(agent)
        return agent

    def _check_ids_and_num(
        self, num: Optional[int], unique_ids: Optional[UniqueIDs]
    ) -> Tuple[int, UniqueIDs]:
        # Checking unique_ids and num of actors:
        unique_ids = make_list(unique_ids)
        if num is None and not unique_ids:
            unique_ids = [None]
        elif isinstance(num, int) and not unique_ids:
            unique_ids = [None] * num
        elif unique_ids and num is None:
            num = len(unique_ids)
        elif isinstance(num, int) and len(unique_ids) != num:
            raise ValueError("Not matching num of actors and unique_ids.")
        return cast(int, num), cast(List, unique_ids)

    def register(self, actor_cls: Type[Actor]) -> None:
        """Registers a new breed of actors.

        Parameters:
            actor_cls:
                The class of the actor to register.
                It can be a single class, or an iterable of classes.
                Once a breed is registered,
                it will be added to all the containers of the model globally.
                It means, it's not necessary to register the same breed again.

        Raises:
            ValueError:
                If the breed is already registered.
        """
        breed = actor_cls.breed
        if breed in self._model.breeds:
            raise ValueError(f"{breed} is already registered.")
        setattr(self._model, "breeds", actor_cls)
        self._unique_ids[breed] = 0

    def new(
        self,
        breed_cls: Type[Actor] = Actor,
        num: Optional[int] = None,
        singleton: bool = False,
        unique_ids: Optional[UniqueIDs] = None,
        **kwargs: Any,
    ) -> Union[Actor, ActorsList[Actor]]:
        """Create one or more actors of the given breed class.

        Parameters:
            breed_cls:
                The breed class of the actor(s) to create.
            num:
                The number of actors to create. Defaults to 1.
            singleton (bool, optional):
                Whether to create a singleton actor. Defaults to False.
            **kwargs:
                Additional keyword arguments to pass to the actor constructor.

        Returns:
            The created actor(s).

        Example:
            ```python
            from abses import Actor, MainModel
            model = MainModel()
            actor = model.agents.new(singleton=True)
            >>> type(actor)
            >>> Actor

            actors = model.agents.new(singleton=False)
            >>> type(actors)
            >>> ActorsList
            ```
        """
        # check if the breed class is registered, if not, register it.
        self.check_registration(breed_cls, register=True)
        num, unique_ids = self._check_ids_and_num(num, unique_ids)

        # create actors.
        objs = []
        for unique_id in unique_ids:
            agent = self._new_one(
                unique_id=unique_id, agent_cls=breed_cls, **kwargs
            )
            objs.append(agent)
        # return the created actor(s).
        actors_list: ActorsList[Actor] = ActorsList(
            model=self.model, objs=objs
        )
        logger.debug(f"{self} created {num} {breed_cls.__name__}.")
        return cast(Actor, actors_list.item()) if singleton else actors_list

    def new_from_graph(
        self,
        graph: "nx.Graph",
        link_name: str,
        actor_cls: Type[Actor] = Actor,
        **kwargs,
    ):
        """Create a set of new agents from networkx graph."""
        self.check_registration(actor_cls, register=True)
        actors = []
        mapping = {}
        for node, attr in graph.nodes(data=True):
            unique_id = get_node_unique_id(node=node)
            actor = self._new_one(unique_id, agent_cls=actor_cls, **attr)
            actors.append(actor)
            mapping[unique_id] = actor
        self.model.human.add_links_from_graph(
            graph, link_name=link_name, mapping_dict=mapping, **kwargs
        )
        return ActorsList(model=self.model, objs=actors)

    def new_from_gdf(
        self,
        gdf: gpd.GeoDataFrame,
        unique_id: Optional[str] = None,
        agent_cls: type[Actor] = Actor,
        attrs: IncludeFlag = False,
        **kwargs,
    ) -> ActorsList[Actor]:
        """Create actors from a `geopandas.GeoDataFrame` object.

        Parameters:
            gdf:
                The `geopandas.GeoDataFrame` object to convert.
            unique_id:
                A column name, to be converted to unique index
                of created geo-agents (Social-ecological system Actors).
            agent_cls:
                Agent class to create.

        Raises:
            ValueError:
                If the column specified by `unique_id` is not unique.

        Returns:
            An `ActorsList` with all new created actors stored.
        """
        # 检查创建主体的数据标识是否唯一，若唯一则设置为索引
        if unique_id:
            gdf = gdf.set_index(unique_id, verify_integrity=True)
        # 如果还没有注册这种主体，那么注册一下
        self.check_registration(agent_cls, register=True)
        # 检查坐标参考系是否一致
        self._check_crs(gdf)
        # 看一下哪些属性是需要加入到主体的
        geo_col = gdf.geometry.name
        set_attributes = clean_attrs(gdf.columns, attrs, exclude=geo_col)
        if not isinstance(set_attributes, dict):
            set_attributes = {col: col for col in set_attributes}
        # 创建主体
        agents = []
        for index, row in gdf.iterrows():
            geometry = row[geo_col]
            new_agent = self._new_one(
                geometry=geometry,
                unique_id=index,
                agent_cls=agent_cls,
                **kwargs,
            )
            new_agent.crs = self.crs

            for col, name in set_attributes.items():
                setattr(new_agent, name, row[col])
            agents.append(new_agent)
        # 添加主体到模型容器里
        self.add(agents)
        return ActorsList(model=self.model, objs=agents)


class _CellAgentsContainer(_AgentsContainer):
    """Container for agents located at cells."""

    def __init__(
        self,
        model: MainModel[Any, Any],
        cell: PatchCell,
        max_len: int | None = None,
    ):
        super().__init__(model, max_len)
        self._cell = cell

    def __str__(self) -> str:
        return "CellAgents"

    def _add_one(self, agent: Actor) -> None:
        if agent.on_earth and agent not in self:
            e1 = f"{agent} is on another cell thus cannot be added."
            e2 = "You may use 'actor.move.to()' to change its location."
            e3 = "Or you may use 'actor.move.off()' before adding it."
            raise ABSESpyError(e1 + e2 + e3)
        super()._add_one(agent)
        agent.at = self._cell

    def remove(self, agent: Actor) -> None:
        """Remove the given agent from the cell.
        Generally, it stores all the agents on this cell.
        Therefore, it is not recommended to use this method directly.
        Consider to use `actor.move.off()` to let the actor leave this cell instead.

        Parameters:
            agent:
                The agent (actor) to remove.

        Raises:
            ABSESpyError:
                If the agent is not on this cell.
        """
        if agent.at is not self._cell:
            raise ABSESpyError(f"{agent} is not on this cell.")
        self[agent.breed].remove(agent)
        del agent.at

    def new(
        self,
        breed_cls: Type[Actor] = Actor,
        num: int = 1,
        singleton: bool = False,
        unique_ids: Optional[UniqueIDs] = None,
        **kwargs: Any,
    ) -> Actor | ActorsList:
        """Creates a new actor or a list of actors of the given breed class.
        The created actors are added to both the cell and the model's global container.

        Parameters:
            breed_cls:
                The breed class of the actor(s) to create.
            num:
                The number of actors to create. Defaults to 1.
            singleton:
                Whether to create a singleton actor. Defaults to False.
            **kwargs:
                Additional keyword arguments to pass to the actor constructor.

        Returns:
            The created actor(s).
        """
        # Using model's container to create a list of actors
        new_actors = self.model.agents.new(
            breed_cls, num, singleton, unique_ids, **kwargs
        )
        # move the actors to the cell
        for a in make_list(new_actors):
            a.move.to(self._cell)
        return new_actors
